import re
import sys
import warnings

import click

from easy_sql.sql_linter.sql_linter import SqlLinter


def split_rules_to_list(rule_description: str):
    if rule_description != "":
        return rule_description.split(",")
    else:
        return None


def parse_backend(sql: str):
    sql_lines = sql.split("\n")
    parsed_backend = None
    for line in sql_lines:
        if re.match(r"^-- \s*backend:.*$", line):
            parsed_backend = line[line.index("backend:") + len("backend:") :].strip()
            break

    if parsed_backend is None:
        parsed_backend = "spark"
    return parsed_backend


def lint_process(
    check_sql_file_path: str, exclude: str, include: str, backend: str, easy_sql: bool, config_path: str = None
):
    if not check_sql_file_path.endswith(".sql"):
        warnings.warn("file name:" + check_sql_file_path + " must end with .sql")

    with open(check_sql_file_path, "r") as file:
        sql = file.read()
    sql_linter = SqlLinter(sql, exclude_rules=split_rules_to_list(exclude), include_rules=split_rules_to_list(include))
    backend = backend if backend else parse_backend(sql)
    print("using backend:", backend)
    result = sql_linter.lint(backend, easysql=easy_sql, config_path=config_path)
    fixed = sql_linter.fix(backend, easy_sql=easy_sql, config_path=config_path)

    return result, fixed


def write_out_fixed(check_sql_file_path: str, fixed: str, inplace: bool):
    if inplace:
        write_out_file_path = check_sql_file_path
    else:
        write_out_file_path = check_sql_file_path.replace(".sql", ".fixed.sql")
    with open(write_out_file_path, "w") as file:
        file.write(fixed)


@click.group()
def cli():
    """Check or fix violations in SQL."""
    pass


def fix_process(
    path: str, exclude: str, include: str, backend: str, inplace: bool, easy_sql: bool, config_path: str = None
):
    result, fixed = lint_process(path, exclude, include, backend, easy_sql, config_path=config_path)
    write_out_fixed(path, fixed, inplace)


@cli.command(help="""Fix rule violations in sql""")
@click.option("--path", help="sql file path", required=True, type=str)
@click.option("--config-path", help="sql fluff config file path, must be named .sqlfluff", required=False, type=str)
@click.option("--exclude", help="comma separated rule to be excluded", default="", required=False, type=str)
@click.option("--include", help="comma separated rule to be included", default="", required=False, type=str)
@click.option(
    "--backend",
    help=(
        "backend for this file, "
        "if easy sql it will parse from the sql file if not specify, "
        "if normal sql it will default to spark"
    ),
    default=None,
    required=False,
    type=str,
)
@click.option("--inplace", help="fix file inplace", default=False, required=False, type=bool)
@click.option("--easy_sql", help="easy sql or normal sql", default=True, required=False, type=bool)
def fix(path: str, config_path: str, exclude: str, include: str, backend: str, inplace: bool, easy_sql: bool):
    fix_process(path, exclude, include, backend, inplace, easy_sql, config_path=config_path)


@cli.command(help="""Check rule violations in sql""")
@click.option("--path", help="sql file path", required=True, type=str)
@click.option("--config-path", help="sql fluff config file path, must be named .sqlfluff", required=False, type=str)
@click.option("--exclude", help="comma separated rule to be excluded", default="", required=False, type=str)
@click.option("--include", help="comma separated rule to be included", default="", required=False, type=str)
@click.option(
    "--backend",
    help=(
        "backend for this file, "
        "if easy sql it will parse from the sql file if not specify, "
        "if normal sql it will default to spark"
    ),
    default=None,
    required=False,
    type=str,
)
@click.option("--easy_sql", help="easy sql or normal sql", default=True, required=False, type=bool)
def lint(path: str, config_path: str, exclude: str, include: str, backend: str, easy_sql: bool):
    lint_process(path, exclude, include, backend, easy_sql, config_path=config_path)


if __name__ == "__main__":
    cli.main(sys.argv[1:])



import unittest

from easy_sql.report import EsService, Reporter


class ReporterTest(unittest.TestCase):
    @unittest.skip("integration test")
    def test_should_report_task_result(self):
        reporter = Reporter(EsService("http://testes:9200"))
        reporter.report_task_result("some-task", "some message\nsome other message")



import unittest

from easy_sql.sql_processor.context import CommentSubstitutor


class CommentSubstitutorTest(unittest.TestCase):
    def test_check_if_quote_closed(self):
        # self.assertTrue(CommentSubstitutor.is_quote_closed(''))
        # self.assertTrue(CommentSubstitutor.is_quote_closed('abc'))
        self.assertTrue(CommentSubstitutor.is_quote_closed('""'))
        self.assertTrue(CommentSubstitutor.is_quote_closed('"a"'))
        self.assertTrue(CommentSubstitutor.is_quote_closed('"\'"'))
        self.assertTrue(CommentSubstitutor.is_quote_closed('"\\""'))
        self.assertTrue(CommentSubstitutor.is_quote_closed('"\\\\\\""'))
        self.assertTrue(CommentSubstitutor.is_quote_closed('"\\\\\\""""'))
        self.assertTrue(CommentSubstitutor.is_quote_closed('"\\\\\\"""."'))
        self.assertTrue(CommentSubstitutor.is_quote_closed('"\\\\\\""."."'))
        self.assertTrue(CommentSubstitutor.is_quote_closed('"\\\\\\""\'\''))
        self.assertTrue(CommentSubstitutor.is_quote_closed('"\\\\\\""\'.\''))
        self.assertTrue(CommentSubstitutor.is_quote_closed('"\\\\\\"."\'.\''))
        self.assertFalse(CommentSubstitutor.is_quote_closed('"\\\\\\""\''))
        self.assertFalse(CommentSubstitutor.is_quote_closed('"\\\\\\"".\''))
        self.assertFalse(CommentSubstitutor.is_quote_closed('"\\\\\\"""'))
        self.assertFalse(CommentSubstitutor.is_quote_closed('"\\\\""'))
        self.assertTrue(CommentSubstitutor.is_quote_closed("''"))
        self.assertTrue(CommentSubstitutor.is_quote_closed("'a'"))
        self.assertTrue(CommentSubstitutor.is_quote_closed("'\"'"))
        self.assertTrue(CommentSubstitutor.is_quote_closed("'\\''"))
        self.assertTrue(CommentSubstitutor.is_quote_closed("'\\\\\\''"))
        self.assertFalse(CommentSubstitutor.is_quote_closed("'\\\\\\''\""))
        self.assertFalse(CommentSubstitutor.is_quote_closed("'\\\\\\'''"))
        self.assertFalse(CommentSubstitutor.is_quote_closed("'\\\\\\''.'"))
        self.assertFalse(CommentSubstitutor.is_quote_closed("'\\\\''"))

    def test_should_replace_and_recover_comment(self):
        sql_expr = """
select ${a}, ${b} -- ${a} in comment
, ',-- ' as c -- special comment
, ',--  as c -- something' -- special comment 1
, ",--  as c -- something" -- special comment 2 ${a}
, ",-- ${a}",--  as c -- something' -- special comment 3 ${a}
, ",\\"-- ${a}",--  as c -- something' -- special comment 4 ${a}
, ",-- ${a}\\\\",--  as c -- something' -- special comment 5 ${a}
, ",\\"-- ${a}\\\\",--  as c -- something' -- special comment 6 ${a}
, ",'-- ${a}",--  as c -- something' -- special comment 7 ${a}
, ',"-- ${a}',--  as c -- something' -- special comment 8 ${a}
, ',-- ' as c -- special comment
-- ${a} in comment
   ${a} -- some comment
        """
        sub = CommentSubstitutor()
        sql_expr = sub.substitute(sql_expr)
        print("\n".join(sub.recognized_comment))
        print(sql_expr)
        sql_expr = sql_expr.replace("${a}", "aaa")
        sql_expr = sql_expr.replace("${b}", "bbb\nbbb")
        self.assertEqual(
            sub.recover(sql_expr),
            """
select aaa, bbb
bbb -- ${a} in comment
, ',-- ' as c -- special comment
, ',--  as c -- something' -- special comment 1
, ",--  as c -- something" -- special comment 2 ${a}
, ",-- aaa",--  as c -- something' -- special comment 3 ${a}
, ",\\"-- aaa",--  as c -- something' -- special comment 4 ${a}
, ",-- aaa\\\\",--  as c -- something' -- special comment 5 ${a}
, ",\\"-- aaa\\\\",--  as c -- something' -- special comment 6 ${a}
, ",'-- aaa",--  as c -- something' -- special comment 7 ${a}
, ',"-- aaa',--  as c -- something' -- special comment 8 ${a}
, ',-- ' as c -- special comment
-- ${a} in comment
   aaa -- some comment
        """,
        )



import os
import re
import sys

import click

from .sql_tester import SqlTester, TestCase, WorkPath


@click.group()
def main():
    pass


def create_backend(backend: str, env: str, case: TestCase, work_path: WorkPath):
    if backend == "spark":
        from easy_sql.local_spark import LocalSpark
        from easy_sql.sql_processor.backend import SparkBackend

        custom_conf = {
            "spark.sql.warehouse.dir": f"/tmp/spark-warehouse-{env}",
            "spark.driver.extraJavaOptions": (
                f"-Dderby.system.home=/tmp/spark-metastore-{env} "
                f"-Dderby.stream.error.file=/tmp/spark-metastore-{env}.log"
            ),
        }
        if case.udf_file_paths or case.func_file_paths:
            paths = case.udf_file_paths + case.func_file_paths
            custom_conf["spark.files"] = ",".join([work_path.path(path) for path in paths])
            # need to recreate SparkSession when there's new config
            LocalSpark.stop()

        import shutil

        shutil.rmtree(f"/tmp/spark-warehouse-{env}", True)

        # try get a clean spark environment
        spark = LocalSpark.get(custom_conf)
        backend = SparkBackend(spark)
        backend.clean()
        return backend
    elif backend == "postgres":
        from easy_sql.sql_processor.backend.rdb import RdbBackend

        assert "PG_URL" in os.environ, "Must set PG_URL env var to run an ETL with postgres backend."
        pg_url = os.environ["PG_URL"]
        backend = RdbBackend(pg_url)

        i = 0
        while True:
            try:
                backend.exec_native_sql(f"drop database if exists unittest{i}")
                break
            except Exception as e:
                print(e.args)
                if re.match(r".*psycopg2.errors.ObjectInUse.*", e.args[0]):
                    i += 1
                else:
                    raise e

        backend.exec_native_sql(f"create database unittest{i}")
        backend = RdbBackend(pg_url[: pg_url.rindex("/")] + f"/unittest{i}")
        return backend
    elif backend == "clickhouse":
        from easy_sql.sql_processor.backend.rdb import RdbBackend

        assert "CLICKHOUSE_URL" in os.environ, "Must set CLICKHOUSE_URL env var to run an ETL with clickhouse backend."
        ch_url = os.environ["CLICKHOUSE_URL"]
        backend = RdbBackend(ch_url)
        return backend
    elif backend == "bigquery":
        from easy_sql.sql_processor.backend.rdb import RdbBackend

        backend = RdbBackend(
            "bigquery://", credentials=f"{os.environ.get('HOME', '/tmp')}/.bigquery/credential-test.json"
        )
        return backend
    else:
        raise Exception("unsupported backend: " + backend)


_FILES_ARG_HELP = "Test files (multiple files split by comma): file1,file2,..."
_BACKEND_ARG_HELP = (
    "Which backend to use. The supported backends are: spark postgres bigquery clickhouse.\n"
    "For spark backend, the temporary folder will be /tmp/spark-metastore-{ENV} .\n"
    "For postgres backend, an environment variable named PG_URL must be set.\n"
    "For clickhouse backend, an environment variable named CLICKHOUSE_URL must be set.\n"
    "For bigquery backend, the credential file must be put at ${HOME}/.bigquery/credential-{ENV}.json .\n"
)
_ENV_ARG_HELP = (
    "Environment identifier to indicate which env to run on. Possible values could be: test e2e\n"
    "Used for logging and temporary file or folder creation."
)


@main.command(name="run-test")
@click.option("--files", "-f", type=str, help=_FILES_ARG_HELP)
@click.option("--backend", "-b", type=str, help=_BACKEND_ARG_HELP, required=False, default="spark", show_default=True)
@click.option("--env", "-e", type=str, help=_ENV_ARG_HELP, required=False, default="test", show_default=True)
def run_test(files, backend: str = "spark", env: str = "test"):
    _run_test(files, backend, env)


def _run_test(files, backend: str = "spark", env: str = "test"):
    test_data_files = [os.path.abspath(f.strip()) for f in files.split(",") if f.strip()]
    for f in test_data_files:
        if not os.path.exists(f):
            print(f"[ERROR] File does not exist: `{sys.argv}`")
            sys.exit(1)

    work_dir = os.path.abspath(os.curdir)
    SqlTester(
        env=env, backend_creator=lambda case: create_backend(backend, env, case, WorkPath(work_dir)), work_dir=work_dir
    ).run_tests(test_data_files)


@main.command(name="convert-json")
@click.option("--files", "-f", type=str, help=_FILES_ARG_HELP)
def convert_json(files):
    _convert_json(files)


def _convert_json(files):
    test_data_files = [os.path.abspath(f.strip()) for f in files.split(",") if f.strip()]
    sql_tester = SqlTester(work_dir=os.path.abspath(os.curdir))
    for f in test_data_files:
        sql_tester.convert_cases_to_json(f)


if __name__ == "__main__":
    main()



import unittest

from easy_sql.utils.object_utils import get_attr


class ObjectUtilsTest(unittest.TestCase):
    def test_get_attr(self):
        self.assertEqual(get_attr({}, "a.b.c"), {})
        self.assertEqual(get_attr({"a": {}}, "a.b.c"), {})
        self.assertEqual(get_attr({"a": {"b": {"c": [1, 2, 3]}}}, "a.b.c"), [1, 2, 3])

        self.assertEqual(get_attr({}, "a"), {})
        self.assertEqual(get_attr({"a": ""}, "a"), "")

        self.assertEqual(get_attr({"a": 1}, ""), {"a": 1})


if __name__ == "__main__":
    unittest.main()



import os
import shutil
from typing import Any, Dict

from pyspark.sql import SparkSession


class LocalSpark:
    spark: SparkSession = None
    __conf: Dict = {}

    @staticmethod
    def stop():
        if LocalSpark.spark:
            LocalSpark.spark.stop()
            LocalSpark.spark = None

    @staticmethod
    def get(conf: Dict[str, Any] = None, clean_existing_data: bool = True) -> SparkSession:
        conf = conf or {}
        if LocalSpark.spark is None:
            default_conf = {
                "spark.default.parallelism": 4,
                "hive.exec.dynamic.partition.mode": "nonstrict",
                "spark.sql.warehouse.dir": "/tmp/spark-warehouse-localdw-ut",
                "spark.driver.extraJavaOptions": (
                    "-Dderby.system.home=/tmp/spark-warehouse-metastore-ut "
                    "-Dderby.stream.error.file=/tmp/spark-warehouse-metastore-ut.log"
                ),
            }
            default_conf.update(conf)
            conf = default_conf

            if clean_existing_data:
                # delete old spark warehouse/metastore dir
                print(f"removing dir {conf['spark.sql.warehouse.dir']}")
                shutil.rmtree(conf["spark.sql.warehouse.dir"], ignore_errors=True)
                if "-Dderby.system.home" in conf["spark.driver.extraJavaOptions"]:
                    import re

                    java_options = re.sub(r"\s*=\s*", "=", conf["spark.driver.extraJavaOptions"].strip()).split()
                    for op in java_options:
                        if op.split("=")[0].strip() == "-Dderby.system.home":
                            print(f"removing dir {op.split('=')[1].strip()}")
                            shutil.rmtree(op.split("=")[1].strip(), ignore_errors=True)

            # ensure a local spark with default config
            os.environ["SPARK_CONF_DIR"] = "/tmp/local-spark-conf-ut"
            spark_builder = SparkSession.builder.appName("UnitTest").master("local[4]")

            print("using conf: ", conf)
            for k, v in conf.items():
                spark_builder.config(k, v)
            LocalSpark.spark = spark_builder.enableHiveSupport().getOrCreate()

        spark = LocalSpark.spark
        spark.catalog.clearCache()
        for table in spark.catalog.listTables("default"):
            if table.isTemporary:
                print(f"dropping temp view {table.name}")
                spark.catalog.dropTempView(table.name)

        return LocalSpark.spark



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ..logger import logger

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession


def _exec_sql(spark: SparkSession, sql: str) -> DataFrame:
    logger.info(f"will exec sql: {sql}")
    return spark.sql(sql)


def is_int_type(type_name):
    return any([type_name.startswith(t) for t in ["integer", "long", "decimal", "short"]])


class Column:
    def __init__(self, name: str, value: Any):
        self.name, self.value = name, value


class SqlProcessorException(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class SqlProcessorAssertionError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class VarsReplacer:
    def replace_variables(self, text: str, include_funcs: bool = True) -> str:
        raise NotImplementedError()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union

from ..logger import logger
from .backend import Backend, SparkBackend
from .context import ProcessorContext, TemplatesContext, VarsContext
from .funcs import FuncRunner
from .report import SqlProcessorReporter, StepStatus
from .step import Step, StepFactory

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

    from .common import Column


def extract_funcs_from_pyfile(funcs_py_file):
    import importlib
    import os
    import sys

    sys.path.insert(0, os.path.dirname(funcs_py_file))
    func_mod = importlib.import_module(os.path.basename(funcs_py_file)[:-3])
    funcs = {func: getattr(func_mod, func) for func in dir(func_mod) if callable(getattr(func_mod, func))}
    return funcs


class SqlProcessor:
    def __init__(
        self,
        backend: Union[SparkSession, Backend],
        sql: str,
        extra_cols: List[Column] = None,
        variables: dict = None,
        report_hdfs_path: str = None,
        report_task_id: str = None,
        report_es_url: str = None,
        report_es_index_prefix: str = None,
        scala_udf_initializer: str = None,
        templates: dict = None,
        includes: Dict[str, str] = None,
    ):
        backend = backend if isinstance(backend, (Backend,)) else SparkBackend(spark=backend)
        self.backend = backend
        self.sql = sql

        self.reporter = SqlProcessorReporter(
            report_task_id=report_task_id,
            report_hdfs_path=report_hdfs_path,
            report_es_url=report_es_url,
            report_es_index_prefix=report_es_index_prefix,
        )
        log_var_tmpl_replace = False
        vars_context = VarsContext(debug_log=log_var_tmpl_replace, vars=variables)
        self.func_runner = FuncRunner.create(self.backend)
        vars_context.init(self.func_runner)
        self.context = ProcessorContext(
            vars_context, TemplatesContext(debug_log=log_var_tmpl_replace, templates=templates), extra_cols=extra_cols
        )
        self.step_factory = StepFactory(self.reporter, self.func_runner)

        self.step_list = self.step_factory.create_from_sql(self.sql, includes)
        self.reporter.init(self.step_list)
        self.backend.init_udfs(scala_udf_initializer=scala_udf_initializer)

    @property
    def variables(self) -> Union[str, Any]:
        return self.context.vars_context.vars

    @property
    def templates(self) -> Dict[str, str]:
        return self.context.templates_context.templates

    @property
    def extra_cols(self) -> List[Column]:
        return self.context.extra_cols

    def set_spark_configs(self, configs: Dict[str, str]):
        if self.backend.is_spark_backend:
            self.backend.set_spark_configs(configs)
        else:
            logger.warn(f"ignored set spark configs when backend is of type {type(self.backend)}")

    def register_funcs_from_pyfile(self, funcs_py_file: str):
        funcs = extract_funcs_from_pyfile(funcs_py_file)
        self.register_funcs(funcs)

    def register_funcs(self, funcs: Dict[str, Callable]):
        self.func_runner.register_funcs(funcs)

    def register_udfs_from_pyfile(self, funcs_py_file: str):
        logger.info(f"resolving udfs from pyfile {funcs_py_file}")
        funcs = extract_funcs_from_pyfile(funcs_py_file)
        self.register_udfs(funcs)

    def register_udfs(self, funcs: Dict[str, Callable]):
        self.backend.register_udfs(funcs)

    def set_vars(self, vars: Dict[str, Any]):
        self.context.set_vars(vars)

    def add_vars(self, vars: Dict[str, Any]):
        self.context.add_vars(vars)

    def run_step(self, step: Step, dry_run: bool):
        try:
            # add meta vars to support step information retrieving in functions
            self.context.add_vars({"__step__": step})
            self.context.add_vars({"__context__": self.context})
            self.context.add_vars({"__backend__": self.backend})
            if not step.should_run(self.context):
                self.reporter.collect_report(step, status=StepStatus.SKIPPED)
                return
            self.reporter.collect_report(step, status=StepStatus.RUNNING)
            df = step.read(self.backend, self.context)
            step.write(self.backend, df, self.context, dry_run)
            self.reporter.collect_report(step, status=StepStatus.SUCCEEDED)
        except Exception as e:
            import traceback

            self.reporter.collect_report(step, status=StepStatus.FAILED, message=traceback.format_exc())
            if (
                "__exception_handler__" in self.variables
                and str(self.variables["__exception_handler__"]).upper() != "NULL"
            ):
                func_name: str = self.variables["__exception_handler__"]
                func_name = func_name.replace("{", "${")
                self.func_runner.run_func(func_name, self.context.vars_context)(e)
            else:
                raise e

    def run(self, dry_run: bool = False):
        try:
            for step in self.step_list:
                self.run_step(step, dry_run)
        finally:
            self.reporter.print_report(True)



import unittest

from pyspark.sql.functions import expr, lit

from easy_sql.base_test import LocalSpark
from easy_sql.sql_processor.backend import SparkTable


class SparkTest(unittest.TestCase):
    def test_with_column(self):
        spark = LocalSpark.get()
        df = spark.sql("select 1 as id")
        # expr('2021-01-01') 效果就是 select 2021-01-01，少了引号，解析就会出现奇怪的结果
        self.assertNotEqual(
            SparkTable(df).with_column("data_date", "2021-01-01").df.select("data_date").limit(1).collect(),
            [("2021-01-01",)],
        )
        # expr("'2021-01-01'") 效果就是 select '2021-01-01'，结果正确
        self.assertEqual(
            SparkTable(df).with_column("data_date", "'2021-01-01'").df.select("data_date").limit(1).collect(),
            [("2021-01-01",)],
        )
        # 可以直接传入 Column
        self.assertEqual(
            SparkTable(df).with_column("data_date", lit("2021-01-01")).df.select("data_date").limit(1).collect(),
            [("2021-01-01",)],
        )
        self.assertEqual(SparkTable(df).with_column("flag", "1==2").df.select("flag").limit(1).collect(), [(False,)])
        self.assertEqual(
            SparkTable(df).with_column("flag", expr("1==2")).df.select("flag").limit(1).collect(), [(False,)]
        )


if __name__ == "__main__":
    unittest.main()



import unittest

from .base_test import TEST_CH_URL, TEST_PG_URL
from .sql_processor.backend import Backend, SparkBackend
from .sql_processor_debugger import SqlProcessorDebugger


class SqlProcessorDebuggerTest(unittest.TestCase):
    def test_process_sql_debugger_spark(self):
        from .base_test import LocalSpark

        spark = LocalSpark.get()
        df = spark.createDataFrame([(1, "a"), (2, "b"), (3, "c"), (4, "d"), (5, "e")], ["id", "type"])
        df.createOrReplaceTempView("target")
        self.run_test_process_sql_debugger(SparkBackend(spark))

    def test_process_sql_debugger_pg(self):
        from easy_sql.sql_processor.backend.postgres import PostgresBackend, _exec_sql

        pg = PostgresBackend(TEST_PG_URL)
        _exec_sql(pg.conn, "drop schema if exists t cascade")
        _exec_sql(pg.conn, "create schema t")
        _exec_sql(pg.conn, "create table t.target(id int, type varchar(100))")
        _exec_sql(pg.conn, "insert into t.target values(1, 'a'), (2, 'b'), (3, 'c'), (4, 'd'), (5, 'e')")
        pg.create_temp_table(pg.exec_sql("select * from t.target"), "target")
        self.run_test_process_sql_debugger(pg)

    def test_process_sql_debugger_ch(self):
        from easy_sql.sql_processor.backend.rdb import RdbBackend, _exec_sql

        backend = RdbBackend(TEST_CH_URL)
        _exec_sql(backend.conn, "drop database if exists t")
        _exec_sql(backend.conn, "create database t")
        _exec_sql(backend.conn, "create table t.target(id int, type String) engine=MergeTree order by tuple()")
        _exec_sql(backend.conn, "insert into t.target values(1, 'a'), (2, 'b'), (3, 'c'), (4, 'd'), (5, 'e')")
        backend.create_temp_table(backend.exec_sql("select * from t.target"), "target")
        self.run_test_process_sql_debugger(backend)

    def run_test_process_sql_debugger(self, backend: Backend):
        sql_content = """
-- target=template.test_a
select
    *
from target
where
    type = 'a';
-- target=variables
select 1 as a
-- target=variables
select '${f2(1, 2)}' as b
-- target=log.b
select '${b}' as b
-- target=log.test_log
select 1 as t
-- target=check.test_check
select 1 as actual, 1 as expected
-- target=check.check(1, 2), if=f1(1, ${a})
-- target=template.test_b
-- for ch, we must use sub query,
-- since `#{type} as type` in select expression conflicts with `target.type` in where expression
select
    id as id,
    #{type} as type
from (
    select
        id as id
    from target
    where
        type = #{type}
) t;
-- target=template.test_c
select
    id as id,
    #{type2} as type
from (
    select
        id as id
    from target
    where
        type = #{type1}
) t;

-- target=temp.test_result
select * from (
    @{test_a()}
) t -- comment
--
union all
select * from (
    @{test_b(type='b')}
) t1
union all
select * from (
    @{test_c(type1='c', type2='${c}')}
) t2
        """

        py_funcs_content = """

check = lambda a, b: a and b

f1 = lambda a, b: a and b

f2 = lambda a, b: a + b
        """
        with open("/tmp/test_sql_processor_debugger.sql", "w") as f:
            f.write(sql_content)
        with open("/tmp/test_sql_processor_debugger_funcs.py", "w") as f:
            f.write(py_funcs_content)
        variables = {"c": "c"}
        debugger = SqlProcessorDebugger(
            "/tmp/test_sql_processor_debugger.sql",
            backend,
            variables,
            {},
            "/tmp/test_sql_processor_debugger_funcs.py",
            [],
        )

        debugger.print_steps()
        debugger.report()
        self.assertTrue(len(debugger.tempviews) in [1, 2])

        self.assertEqual(debugger.current_step_no, None)
        self.assertEqual(debugger.next_step_no, 1)
        self.assertEqual(debugger.last_step_no, None)
        self.assertEqual(debugger.left_step_count, len(debugger.steps))
        self.assertFalse(debugger.is_started)
        self.assertFalse(debugger.is_finished)
        self.assertIsNone(debugger.current_step)
        self.assertIsNone(debugger.last_step)

        debugger.step_on()
        self.assertEqual(debugger.current_step_no, 1)
        self.assertEqual(debugger.next_step_no, 2)
        self.assertEqual(debugger.last_step_no, None)
        self.assertEqual(debugger.left_step_count, len(debugger.steps) - 1)
        self.assertTrue(debugger.is_started)
        self.assertFalse(debugger.is_finished)
        self.assertEqual(debugger.current_step, debugger.steps[0])
        debugger.step_to(1)  # will print info, but do nothing
        debugger.run_to(1)  # will print info, but do nothing

        debugger.step_to(3)
        self.assertEqual(debugger.current_step_no, 3)
        self.assertEqual(debugger.next_step_no, 4)
        self.assertEqual(debugger.last_step_no, 2)

        debugger.run()
        self.assertEqual(debugger.current_step_no, len(debugger.steps))
        self.assertEqual(debugger.current_step, debugger.steps[-1])
        self.assertIsNone(debugger.next_step_no)
        self.assertIsNone(debugger.next_step)
        self.assertEqual(debugger.last_step_no, len(debugger.steps) - 1)
        self.assertEqual(debugger.left_step_count, 0)
        self.assertEqual(debugger.last_step, debugger.steps[-2])

        self.assertListEqual(
            debugger.sql("select * from test_result order by id").collect(),
            [
                (1, "a"),
                (2, "b"),
                (3, "c"),
            ],
        )

        self.assertGreater(len(debugger.vars), 0)
        vars_count = len(debugger.vars)
        debugger.add_vars(None)  # will print info and do nothing
        debugger.add_vars({"__some_var": 1})
        self.assertEqual(len(debugger.vars), vars_count + 1)
        debugger.set_vars({"__some_var": 1})
        self.assertEqual(len(debugger.vars), vars_count + 1)  # vars will not change when finished
        self.assertEqual(len(debugger.initial_vars), 1)
        debugger.set_vars(variables)
        self.assertGreater(len(debugger.templates), 0)
        debugger.showdf("test_result")

        debugger.step_to(-1)  # will print info and do nothing

        debugger.restart()
        self.assertEqual(debugger.current_step_no, None)
        self.assertEqual(debugger.next_step_no, 1)
        self.assertEqual(debugger.last_step_no, None)
        self.assertEqual(debugger.left_step_count, len(debugger.steps))
        self.assertTrue(len(debugger.tempviews) in [1, 2])

        debugger.run()
        self.assertEqual(debugger.current_step_no, len(debugger.steps))
        self.assertEqual(debugger.next_step_no, None)
        self.assertEqual(debugger.last_step_no, len(debugger.steps) - 1)
        self.assertEqual(debugger.left_step_count, 0)

        self.assertListEqual(
            debugger.sql("select * from test_result order by id").collect(),
            [
                (1, "a"),
                (2, "b"),
                (3, "c"),
            ],
        )
        debugger.report()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from .sql_processor import Column, SqlProcessor, Step

__all__ = ["SqlProcessorDebugger"]

from .sql_processor.backend import Backend, SparkBackend

if TYPE_CHECKING:
    from pyspark.sql import SparkSession


class SqlProcessorDebugger:
    def __init__(
        self,
        sql_file_path: str,
        backend: Union[SparkSession, Backend],
        vars: Dict[str, Any] = None,
        funcs: Dict[str, Any] = None,
        funcs_py_file: str = None,
        extra_cols: List[Column] = None,
        udf_py_file: str = None,
        scala_udf_initializer: str = None,
        templates: Dict[str, Any] = None,
    ):
        backend = backend if isinstance(backend, (Backend,)) else SparkBackend(spark=backend)
        self.udf_py_file = udf_py_file
        self.sql_file_path = sql_file_path
        self.scala_udf_initializer = scala_udf_initializer
        self.initial_vars, self.initial_funcs, self.funcs_py_file, self.initial_extra_cols, self.initial_templates = (
            vars or {},
            funcs or {},
            funcs_py_file,
            extra_cols or [],
            templates or {},
        )
        self.backend = backend
        self.sql_processor = self._create_sql_processor()
        self.steps = self.sql_processor.step_list
        self._current_step_index = -1
        self.initial_temp_views = self.tempviews

    def _create_sql_processor(self) -> SqlProcessor:
        import copy

        with open(self.sql_file_path, "r") as f:
            sql = f.read()
        sql_processor = SqlProcessor(
            self.backend,
            sql,
            extra_cols=copy.deepcopy(self.initial_extra_cols),
            variables=copy.deepcopy(self.initial_vars),
            scala_udf_initializer=self.scala_udf_initializer,
            templates=copy.deepcopy(self.initial_templates),
        )
        if self.initial_funcs:
            sql_processor.register_funcs(self.initial_funcs)
        if self.funcs_py_file:
            sql_processor.register_funcs_from_pyfile(self.funcs_py_file)
        if self.udf_py_file:
            sql_processor.register_udfs_from_pyfile(self.udf_py_file)
        return sql_processor

    @property
    def is_started(self) -> bool:
        return self._current_step_index > -1

    @property
    def is_inprogress(self) -> bool:
        return -1 < self._current_step_index < len(self.steps) - 1

    @property
    def is_finished(self):
        return self._current_step_index == len(self.steps) - 1

    @property
    def current_step(self) -> Optional[Step]:
        if self._current_step_index < len(self.steps) and self._current_step_index != -1:
            return self.steps[self._current_step_index]
        return None

    @property
    def current_step_no(self) -> Optional[int]:
        if self._current_step_index < len(self.steps) and self._current_step_index != -1:
            return self._current_step_index + 1
        if self._current_step_index == -1:
            print("Not started yet! No current step number right now.")
        else:
            print("Already finished! No current step number right now.")

    @property
    def next_step(self) -> Optional[Step]:
        if self._current_step_index < len(self.steps) - 1:
            return self.steps[self._current_step_index + 1]
        return None

    @property
    def next_step_no(self) -> Optional[int]:
        step_index = self._current_step_index + 1
        if step_index < len(self.steps) and step_index != -1:
            return step_index + 1
        else:
            print("Already finished! No next step number right now.")

    @property
    def last_step(self) -> Optional[Step]:
        if self._current_step_index > 0:
            return self.steps[self._current_step_index - 1]
        return None

    @property
    def last_step_no(self) -> Optional[int]:
        step_index = self._current_step_index - 1
        if step_index == -2:
            print("Not started yet! No last step number right now.")
            return None
        if step_index < len(self.steps) and step_index != -1:
            return step_index + 1

    @property
    def left_step_count(self) -> int:
        return len(self.steps) - 1 - self._current_step_index

    @property
    def vars(self) -> Dict[str, Any]:
        return self.sql_processor.variables

    def add_vars(self, vars: Dict[str, Any]):
        if vars is None or not isinstance(vars, dict):
            print("Vars must be a non-empty dict. Will do nothing!")
            return
        self.sql_processor.add_vars(vars)
        self.initial_vars.update(vars)

    def set_vars(self, vars: Dict[str, Any]):
        if vars is None or not isinstance(vars, dict):
            print("Vars must be a non-empty dict. Will do nothing!")
            return
        if self.is_inprogress:
            self.sql_processor.set_vars(vars)
        self.initial_vars = vars

    def set_spark_configs(self, configs: Dict[str, str]):
        self.sql_processor.set_spark_configs(configs)

    @property
    def templates(self) -> Dict[str, str]:
        return self.sql_processor.templates

    @property
    def tempviews(self) -> List[str]:
        return self.backend.temp_tables()

    def refresh_initial_tempview(self):
        self.initial_temp_views = self.tempviews

    def native_sql(self, sql: str) -> Any:
        return self.backend.exec_native_sql(sql)

    def sql(self, sql: str) -> Any:
        return self.backend.exec_sql(sql)

    def showdf(self, table_name: str):
        self.sql(f"select * from {table_name}").show()

    def step(self, step_no: int) -> Optional[Step]:
        return self.steps[step_no - 1] if 1 <= step_no <= len(self.steps) else None

    def print_steps(self):
        for i in range(len(self.steps)):
            print(f"Step {i + 1}: {str(self.step(i + 1))}")

    def step_on(self):
        if self.step(self._current_step_index + 1 + 1):
            self.sql_processor.run_step(self.steps[self._current_step_index + 1], True)
            self._current_step_index += 1
        else:
            print("Process already ended! Nothing to run!")

    def step_to(self, step_no: int):
        if step_no <= 0 or step_no > len(self.steps):
            print(f"step_index must be from [1...{len(self.steps)}], got {step_no}. Will not run anything!")
            return
        step_index_0_based = step_no - 1
        if step_index_0_based <= self._current_step_index:
            print(f"We are at step {self._current_step_index + 1} now. Nothing to run!")
            return
        while self._current_step_index < step_index_0_based:
            self.step_on()

    def run(self):
        for _ in range(self.left_step_count):
            self.step_on()

    def run_to(self, step_no: int):
        return self.step_to(step_no)

    def restart(self):
        self.backend.clear_cache()
        self.backend.clear_temp_tables(exclude=self.initial_temp_views)
        self.sql_processor = self._create_sql_processor()
        self.steps = self.sql_processor.step_list
        self._current_step_index = -1

    def report(self, verbose: bool = True):
        self.sql_processor.reporter.print_report(verbose=verbose)



from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

if TYPE_CHECKING:
    from sqlalchemy.engine import ResultProxy
    from sqlalchemy.types import TypeEngine

    from ..base import Partition

__all__ = ["SqlExpr", "SqlDialect"]

from ...common import SqlProcessorAssertionError


class SqlExpr:
    def __init__(
        self,
        value_to_sql_expr: Callable[[Any], Optional[str]] = None,
        column_sql_type_converter: Callable[[str, str, TypeEngine], Optional[str]] = None,
        partition_col_converter: Callable[[str], str] = None,
        partition_value_converter: Callable[[str, str], Any] = None,
        partition_expr: Callable[[str, str], str] = None,
    ):
        self.value_to_sql_expr = value_to_sql_expr
        self.column_sql_type_converter = column_sql_type_converter
        self.partition_col_converter = partition_col_converter
        self.partition_value_converter = partition_value_converter
        self.partition_expr = partition_expr

    def convert_partition_col(self, partition_col: str) -> str:
        if self.partition_col_converter:
            return self.partition_col_converter(partition_col)
        return partition_col

    def bigquery_partition_expr(self, partition_col: str) -> str:
        if self.partition_expr:
            return self.partition_expr("bigquery", partition_col)
        return partition_col

    def convert_partition_value(self, partition_col: str, value: str) -> Any:
        if self.partition_value_converter:
            return self.partition_value_converter(partition_col, value)
        return value

    def for_value(self, value: Union[str, int, float, datetime, date]) -> str:
        if self.value_to_sql_expr:
            sql_expr = self.value_to_sql_expr(value)
            if sql_expr is not None:
                return sql_expr

        if not isinstance(value, (str, int, float, datetime, date)):
            raise SqlProcessorAssertionError(
                "when create new columns, the current supported value types are [str, int, float, datetime, date],"
                f" found: value={value}, type={type(value)}"
            )
        if isinstance(value, (str,)):
            return f"'{value}'"
        elif isinstance(
            value,
            (
                int,
                float,
            ),
        ):
            return f"{value}"
        elif isinstance(value, (datetime,)):
            return f"cast('{value.strftime('%Y-%m-%d %H:%M:%S')}' as timestamp)"
        elif isinstance(value, (date,)):
            return f"cast ('{value.strftime('%Y-%m-%d')}' as date)"
        else:
            raise SqlProcessorAssertionError(f"value of type {type(value)} not supported yet!")

    def for_bigquery_type(self, col_name: str, col_type: Union[str, TypeEngine]) -> str:
        if self.column_sql_type_converter:
            converted_col_type = self.column_sql_type_converter("bigquery", col_name, col_type)
            if converted_col_type is not None:
                return converted_col_type

        if str(col_type.__class__) == "<class 'str'>":
            return col_type

        import sqlalchemy

        if isinstance(col_type, (sqlalchemy.FLOAT, sqlalchemy.Float)):
            return "FLOAT64"
        elif isinstance(col_type, sqlalchemy.VARCHAR):
            return "STRING"
        else:
            return col_type.__class__.__name__


class SqlDialect:
    def __init__(self, sql_expr: SqlExpr):
        self.sql_expr = sql_expr

    def create_partition_automatically(self) -> bool:
        raise NotImplementedError()

    def support_static_partition(self) -> bool:
        raise NotImplementedError()

    def support_native_partition(self) -> bool:
        raise NotImplementedError()

    def support_move_individual_partition(self) -> bool:
        raise NotImplementedError()

    def create_db_sql(self, db: str) -> str:
        raise NotImplementedError()

    def use_db_sql(self, db: str) -> str:
        raise NotImplementedError()

    def drop_db_sql(self, db: str) -> Union[str, List[str]]:
        raise NotImplementedError()

    def rename_table_sql(self, from_table: str, to_table: str) -> str:
        raise NotImplementedError()

    def rename_table_db_sql(self, table_name: str, schema: str):
        raise NotImplementedError()

    def get_tables_sql(self, db) -> str:
        raise NotImplementedError()

    def get_dbs_sql(self) -> str:
        raise NotImplementedError()

    def create_table_sql(self, table_name: str, select_sql: str) -> str:
        raise NotImplementedError()

    def rename_view_sql(self, from_table: str, to_table: str) -> str:
        raise NotImplementedError()

    def drop_view_sql(self, table: str) -> str:
        raise NotImplementedError()

    def create_view_sql(self, table_name: str, select_sql: str) -> str:
        raise NotImplementedError()

    def delete_partition_sql(self, table_name, partitions: List[Partition]) -> Union[str, List[str]]:
        raise NotImplementedError()

    def native_partitions_sql(self, table_name: str) -> Tuple[str, Callable[[ResultProxy], List[str]]]:
        raise NotImplementedError()

    def create_table_with_partitions_sql(self, table_name: str, cols: List[Dict], partitions: List[Partition]):
        raise NotImplementedError()

    def create_partitions_with_data_sqls(
        self, source_table_name: str, target_table_name: str, col_names: List[str], partitions: List[List[Partition]]
    ):
        raise NotImplementedError()

    def create_partition_sql(
        self, target_table_name: str, partitions: List[Partition], if_not_exists: bool = False
    ) -> str:
        raise NotImplementedError()

    def convert_pt_col_expr(self, all_cols: List[str], partition_cols: List[str]) -> List[str]:
        if len(partition_cols) == 0:
            return all_cols
        else:
            if len(partition_cols) > 1:
                raise SqlProcessorAssertionError(
                    f"Only single-column partitioning is supported! found: {partition_cols}"
                )
            # the format of pt_col may be :{pt_col}, which is transferred by method create_table_with_data
            return [
                col
                if col not in [partition_cols[0], f":{partition_cols[0]}"]
                else self.sql_expr.convert_partition_col(col)
                for col in all_cols
            ]

    def insert_data_sql(
        self, table_name: str, col_names_expr: str, select_sql: str, partitions: List[Partition]
    ) -> Union[str, List[str]]:
        raise NotImplementedError()

    def move_data_sql(self, target_table_name: str, temp_table_name: str, partitions: List[Partition]) -> List[str]:
        raise NotImplementedError()

    def drop_table_sql(self, table: str) -> str:
        raise NotImplementedError()

    def create_pt_meta_table_sql(self, db: str) -> str:
        raise NotImplementedError()

    def insert_pt_metadata_sql(self, table_name: str, partitions: List[Partition]) -> str:
        raise NotImplementedError()

    def delete_pt_metadata_sql(self, table_name: str, partitions: List[Partition]) -> str:
        raise NotImplementedError()

    def create_table_like_sql(self, target_table_name: str, source_table_name: str, partitions: List[Partition]) -> str:
        raise NotImplementedError()



from easy_sql.sql_processor.common import Column, SqlProcessorException
from easy_sql.sql_processor.context import VarsContext
from easy_sql.sql_processor.funcs import FuncRunner
from easy_sql.sql_processor.report import SqlProcessorReporter, StepReport, StepStatus
from easy_sql.sql_processor.sql_processor import SqlProcessor
from easy_sql.sql_processor.step import Step, StepConfig, StepType

__all__ = [
    "Column",
    "SqlProcessorException",
    "StepConfig",
    "StepType",
    "VarsContext",
    "FuncRunner",
    "Step",
    "StepStatus",
    "StepReport",
    "SqlProcessorReporter",
    "SqlProcessor",
]



from datetime import datetime
from typing import List, Optional

from .step import ReportCollector, Step


class StepStatus:
    NOT_STARTED = "NOT_STARTED"
    SKIPPED = "SKIPPED"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCEEDED = "SUCCEEDED"


class StepReport:
    def __init__(self, step: Step):
        self.step = step
        self.status = StepStatus.NOT_STARTED
        self.start_time: Optional[datetime] = None
        self.end_time: Optional[datetime] = None
        self.execution_time = 0
        self.messages = []

    def update(self, status: str = None, message: str = None) -> "StepReport":
        if status is not None:
            if status == StepStatus.RUNNING:
                self.start_time = datetime.now()
            if status in [StepStatus.FAILED, StepStatus.SUCCEEDED]:
                self.end_time = datetime.now()
                start_time = self.start_time or datetime.now()
                self.execution_time = (datetime.now() - start_time).total_seconds()
            self.status = status
        if message is not None and message != "":
            self.messages.append(message)
        return self

    def report_as_text(self, total_execution_time: float, verbose: bool = False):
        start_time_str = self.start_time.strftime("%Y-%m-%d %H:%M:%S") if self.start_time is not None else ""
        end_time_str = self.end_time.strftime("%Y-%m-%d %H:%M:%S") if self.end_time is not None else ""
        verbose_parts = "\n" + f"sql: {self.step.select_sql}" if verbose else ""
        return f"""
===================== REPORT FOR {self.step.id} ==================
config: {self.step.target_config} {verbose_parts}
status: {self.status}
start time: {start_time_str}, end time: {end_time_str}, execution time: {self.execution_time}s - {self.execution_time / total_execution_time * 100:.2f}%
messages:
""" + "\n".join(
            self.messages
        )


class SqlProcessorReporter(ReportCollector):
    def __init__(
        self,
        report_task_id: str,
        report_hdfs_path: str = None,
        report_es_url: str = None,
        report_es_index_prefix: str = None,
    ):
        self.report_task_id = report_task_id
        self.report_hdfs_path = report_hdfs_path
        self.report_es_url = report_es_url
        self.report_es_index_prefix = report_es_index_prefix
        self._steps = None
        self.step_reports = None
        self.start_time = datetime.now()

    def init(self, steps: List[Step]):
        self._steps = steps
        self.step_reports: dict[str, StepReport] = {step.id: StepReport(step) for step in steps}

    def collect_report(self, step: Step, status: str = None, message: str = None):
        report = self.step_reports[step.id]
        if status is not None:
            report.update(status=status)
        if message is not None:
            report.update(message=message)

    def print_report(self, verbose: bool = False):
        report = self.get_report(verbose)
        self._report_to_stdout(report)
        if self.report_hdfs_path is not None:
            self._report_to_hdfs(report)
        if self.report_es_url is not None:
            self._report_to_es(report)

    def _report_to_es(self, report):
        from easy_sql.report import EsService, Reporter

        Reporter(EsService(self.report_es_url), index_prefix=self.report_es_index_prefix).report_task_result(
            task_id=self.report_task_id, report=report
        )

    def _report_to_stdout(self, report):
        print(report)

    def _report_to_hdfs(self, report):
        import hashlib
        import os

        file_name = os.path.basename(self.report_hdfs_path)
        md5 = hashlib.md5()
        md5.update(file_name.encode("utf8"))
        folder = md5.hexdigest()[:2]
        local_tmp_folder = f"/tmp/sql_processor_reports/{folder}/"
        local_tmp_file = f"{local_tmp_folder}/{file_name}"

        os.makedirs(local_tmp_folder, exist_ok=True)
        with open(local_tmp_file, "w") as f:
            f.write(report)

        import subprocess

        subprocess.run(
            ["bash", "-c", f"hdfs dfs -rm {self.report_hdfs_path} || exit 0"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        result = subprocess.run(
            [
                "bash",
                "-c",
                f"hdfs dfs -mkdir -p {os.path.basename(self.report_hdfs_path)} && "
                f"hdfs dfs -put {local_tmp_file} {self.report_hdfs_path}",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        if result.returncode != 0:
            raise Exception(f"upload file {local_tmp_file} to hdfs failed, result code: {result.returncode}")

    def get_report(self, verbose: bool = False) -> str:
        report = []
        total_execution_seconds = (datetime.now() - self.start_time).total_seconds()
        for step in self._steps:
            report.append(self.step_reports[step.id].report_as_text(total_execution_seconds, verbose))
        report.append(f"\ntotal execution time: {total_execution_seconds}s")
        return "\n".join(report)



from __future__ import annotations

import time
from enum import Enum
from random import random
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

from ...logger import logger
from .base import Backend, Row, SaveMode, Table

if TYPE_CHECKING:
    from .base import TableMeta

# TODO: SqlExpr should be a common class
from .rdb import SqlExpr


def _exec_sql(conn, sql: str):
    logger.info(f"will exec sql: {sql}")
    conn.execute_sql(sql)


class PartitionMode(Enum):
    ALL_DYNAMIC = (0,)
    ALL_STATIC = (1,)
    HYBRID = 2


class MaxComputeRow(Row):
    def __init__(self, schema, values: Optional[Tuple]):
        from odps.types import Record

        self.row = Record(schema=schema, values=values)
        self.columns = [c.name for c in schema.columns]
        self.types = [c.type.name for c in schema.columns]
        self.values = self.row.values

    @staticmethod
    def from_schema_meta(
        cols: List[str],
        types: List[str],
        values: Optional[Tuple],
        pt_cols: List[str] = None,
        pt_types: List[str] = None,
    ):
        from odps.types import OdpsSchema

        schema = OdpsSchema.from_lists(cols, types, partition_names=pt_cols, partition_types=pt_types)
        return MaxComputeRow(schema=schema, values=values)

    def as_dict(self):
        return dict(self.row)

    def as_tuple(self) -> Tuple:
        return self.row

    def __eq__(self, other):
        if not isinstance(other, MaxComputeRow):
            return False
        return self.columns == other.columns and self.values == other.values

    def __str__(self):
        return str(self.row)[12:]

    def __getitem__(self, i):
        return self.row[i]

    def __repr__(self):
        return self.row.__repr__()


class MaxComputeTable(Table):
    def __init__(self, backend: MaxComputeBackend, sql: str, table_name: str = None):
        self.sql = sql
        self.backend = backend
        # TODO: consider and the job id, for batch delete views
        # TODO: eg. backend.job_id, as to the part of temp table name
        # TODO: table_name need with suffix to avoid parallel run
        self.table_name = table_name or f"t_{int(time.mktime(time.gmtime()))}_{int(random() * 10000):04d}"
        self.backend.conn.execute_sql(f"create or replace view {self.table_name} as {sql}")
        # The MaxCompute view can not be read when using table frame, must transform to dataframe
        self.df = self.backend.get_table(self.table_name).to_df()
        self.backend.append_temp_view(self.table_name)
        from odps.types import OdpsSchema

        self.schema = OdpsSchema.from_lists(self.df.data.schema.names, self.df.data.schema.types)

    @staticmethod
    def from_table_meta(backend, table_meta: TableMeta):
        table = MaxComputeTable(backend, f"select * from {table_meta.table_name}")
        if table_meta.has_partitions():
            for pt in table_meta.partitions:
                if pt.field not in table.field_names():
                    table = table.with_column(pt.field, pt.value)
                else:
                    if pt.value is not None:
                        raise Exception(
                            f"partition column already exists in table {table_meta.table_name}, "
                            f"but right now we provided a new value {pt.value} for partition column {pt.field}"
                        )
        return table

    def is_empty(self) -> bool:
        return self.count() == 0

    def field_names(self) -> List[str]:
        return self.schema.names

    def field_types(self) -> List[str]:
        return self.schema.types

    def first(self) -> MaxComputeRow:
        import numpy

        row = [x.values for x in self.df.head(1)]
        value = row[0] if row else None
        value = [bool(v) if isinstance(v, numpy.bool_) else v for v in value]
        return MaxComputeRow(schema=self.schema, values=tuple(value))

    def limit(self, count: int) -> MaxComputeTable:
        return MaxComputeTable(self.backend, f"select * from {self.table_name} limit {count}")

    def with_column(self, name: str, value: any) -> MaxComputeTable:
        value_expr = self.backend.sql_expr.for_value(value).replace("''", "'")
        return MaxComputeTable(self.backend, f"select *, {value_expr} as {name} from {self.table_name}")

    def collect(self, count: int = 1000) -> List[Row]:
        return [MaxComputeRow(schema=self.schema, values=tuple(r.values)) for r in self.df.head(count)]

    def show(self, count: int = 20):
        print("\t".join(self.field_names()))
        for record in self.df.head(20):
            print("\t".join([f"{x}" for x in record.values]))

    def count(self) -> int:
        return self.df.count().execute()

    def __repr__(self):
        return self.table_name

    # TODO: common code with Table.__partition_expr__, consider add a util class
    def __partition_expr__(self, target_table_meta: TableMeta, partition_mode: PartitionMode) -> str:
        if not target_table_meta.partitions:
            return ""

        dynamic_partitions = [p for p in target_table_meta.partitions if not p.value]
        static_partitions = [p for p in target_table_meta.partitions if p.value]

        if partition_mode == PartitionMode.ALL_DYNAMIC:
            fields = [f"{p.field}" for p in target_table_meta.partitions]
        elif partition_mode == PartitionMode.HYBRID:
            fields = [f"{p.name}='{p.value}'" for p in static_partitions] + [f"{p.field}" for p in dynamic_partitions]
        else:
            if dynamic_partitions:
                raise Exception(
                    "In ALL_STATIC partition mode, no dynamic partition is supported. "
                    f"Now there are the following dynamic partitions: {dynamic_partitions}"
                )
            fields = [f"{p.field}='{p.value}'" for p in target_table_meta.partitions]

        return f"partition ({','.join(fields)})"

    def save_to_table(self, target_table_meta: TableMeta, save_mode: SaveMode = SaveMode.overwrite) -> None:
        dynamic_partitions = [p for p in target_table_meta.partitions if not p.value]
        static_partitions = [p for p in target_table_meta.partitions if p.value]
        target_table = self.backend.get_table(target_table_meta.table_name)

        temp_res = self
        if dynamic_partitions:
            for pt in static_partitions:
                temp_res = self.with_column(pt.field, pt.value)
            target_cols = [c.name for c in target_table.schema.columns]
            partition_expr = self.__partition_expr__(target_table_meta, PartitionMode.ALL_DYNAMIC)

        else:
            target_cols = target_table.schema.names
            partition_expr = self.__partition_expr__(target_table_meta, PartitionMode.ALL_STATIC)

        temp_res = self.backend.exec_sql(f"select {', '.join(target_cols)} from {temp_res.table_name}")

        self.backend.exec_native_sql(
            f"""
            insert {'into' if save_mode == SaveMode.append else save_mode.name}
            table {target_table_meta.table_name}
            {partition_expr}
            select * from {temp_res.table_name}
        """
        )


class MaxComputeBackend(Backend):
    def __init__(self, sql_expr: SqlExpr, **kwargs):
        from odps import ODPS

        self.conn = ODPS(**kwargs)
        self.temp_views = []
        self.sql_expr = sql_expr or SqlExpr()

        from odps import options

        # Support timestamp
        options.sql.use_odps2_extension = True
        # TODO: Support full scan for partitioned table. Should forbid?
        options.sql.settings = {"odps.sql.allow.fullscan": True}

    def get_table(self, table_name):
        return self.conn.get_table(table_name)

    def init_udfs(self, *args, **kwargs):
        pass

    def register_udfs(self, funcs: Dict[str, Callable]):
        pass

    def create_empty_table(self):
        # TODO: empty table?
        return MaxComputeTable(self, "select 1 as a")

    def exec_native_sql(self, sql: str) -> Any:
        logger.info(f"will exec sql: {sql}")
        return self.conn.execute_sql(sql)

    def exec_sql(self, sql: str) -> MaxComputeTable:
        # TODO: extract all temp table names and replace as the mapped temp view name
        return MaxComputeTable(self, sql)

    def temp_tables(self) -> List[str]:
        return self.temp_views

    def append_temp_view(self, view_name):
        self.temp_views.append(view_name)

    def clear_cache(self):
        pass

    def clean(self):
        self.clear_temp_tables()

    def _clear_temp_views(self, exclude: List[str] = None):
        # TODO: consider view dependencies
        for temp_view in self.temp_views:
            if exclude and temp_view in exclude:
                continue
            self.conn.execute_sql(f"drop view if exists {temp_view}")

    def clear_temp_tables(self, exclude: List[str] = None):
        self._clear_temp_views(exclude)

    def create_temp_table(self, table: MaxComputeTable, name: str):
        logger.info(f"create_temp_table with: table={table}, name={name}")
        self.conn.execute_sql(f"create or replace view {name} as select * from {table.table_name}")
        self.append_temp_view(name)

    def create_cache_table(self, table: MaxComputeTable, name: str):
        logger.info(f"create_cache_table with: table={table}, name={name}")
        self.conn.execute_sql(f"create or replace view {name} as select * from {table.table_name}")
        self.append_temp_view(name)

    def broadcast_table(self, table: MaxComputeTable, name: str):
        logger.info(f"broadcast_table with: table={table}, name={name}")
        self.conn.execute_sql(f"create or replace view {name} as select * from {table.table_name}")
        self.append_temp_view(name)

    def table_exists(self, table: TableMeta):
        return self.conn.exist_table(table.table_name)

    def refresh_table_partitions(self, table: TableMeta):
        pass

    def save_table(
        self,
        source_table_meta: TableMeta,
        target_table_meta: TableMeta,
        save_mode: SaveMode,
        create_target_table: bool,
    ):
        logger.info(
            f"save table with: source_table={source_table_meta.table_name}, "
            f"target_table={target_table_meta.table_name}, "
            f"save_mode={save_mode}, create_target_table={create_target_table}"
        )
        if not self.table_exists(target_table_meta) and not create_target_table:
            raise Exception(
                f"target table {target_table_meta.table_name} does not exist, "
                "and create_target_table is False, "
                f"cannot save table {source_table_meta.table_name} to {target_table_meta.table_name}"
            )

        source_table = MaxComputeTable.from_table_meta(self, source_table_meta)

        if not self.table_exists(target_table_meta) and create_target_table:
            pt_names = [p.field for p in target_table_meta.partitions] or None
            pt_types = ["string"] * (len(target_table_meta.partitions) + 1) or None

            from odps.types import OdpsSchema

            self.conn.create_table(
                target_table_meta.table_name,
                schema=OdpsSchema.from_lists(
                    names=source_table.schema.names,
                    types=source_table.schema.types,
                    partition_names=pt_names,
                    partition_types=pt_types,
                ),
            )

        source_table.save_to_table(target_table_meta, save_mode)



import json
import unittest
from pathlib import Path
from typing import Tuple

from easy_sql.base_test import TEST_CH_URL, TEST_PG_URL
from easy_sql.local_spark import LocalSpark
from easy_sql.sql_processor import FuncRunner, Step
from easy_sql.sql_processor.backend import Partition, SparkTable
from easy_sql.sql_processor.backend.base import Backend, Col
from easy_sql.sql_processor.backend.rdb import RdbBackend
from easy_sql.sql_processor.context import (
    ProcessorContext,
    TemplatesContext,
    VarsContext,
)
from easy_sql.sql_processor.funcs_common import (
    Alerter,
    AlertFunc,
    ColumnFuncs,
    TableFuncs,
)
from easy_sql.sql_processor.funcs_rdb import PartitionFuncs
from easy_sql.sql_processor.funcs_spark import CacheFuncs, IOFuncs, ParallelismFuncs
from easy_sql.sql_processor.funcs_spark import PartitionFuncs as SparkPartitionFuncs
from easy_sql.sql_processor.step import ReportCollector


class FuncsRdbTest(unittest.TestCase):
    test_table_name = "t.func_test"

    def test_funcs_pg(self):
        from easy_sql.sql_processor.backend.postgres import PostgresBackend

        backend = PostgresBackend(TEST_PG_URL)
        int_type, str_type, pt_type = "int", "text", "text"
        self.run_test(backend, (int_type, str_type, pt_type))

    def test_funcs_ch(self):
        from easy_sql.sql_processor.backend.clickhouse import ChBackend

        backend = ChBackend(TEST_CH_URL)
        int_type, str_type, pt_type = "Nullable(Int32)", "String", "String"
        self.run_test(backend, (int_type, str_type, pt_type))

    def test_funcs_spark(self):
        from easy_sql.sql_processor.backend.spark import SparkBackend

        backend = SparkBackend(LocalSpark.get())
        int_type, str_type, pt_type = "int", "string", "string"
        self.run_test(backend, (int_type, str_type, pt_type))

        cf = CacheFuncs(backend.spark)
        backend.create_cache_table(SparkTable(backend.exec_native_sql(f"select * from {self.test_table_name}")), "tc")
        self.assertEqual(backend.exec_sql("select count(1) from tc").collect()[0][0], 2)
        cf.unpersist("tc")

        pf = ParallelismFuncs(backend.spark)
        self.assertEqual(backend.exec_sql("select count(1) from tc").collect()[0][0], 2)
        pf.coalesce("tc", "1")
        self.assertEqual(backend.exec_sql("select count(1) from tc").collect()[0][0], 2)
        pf.repartition("tc", "2")
        self.assertEqual(backend.exec_sql("select count(1) from tc").collect()[0][0], 2)
        pf.repartition_by_column("tc", "pt")
        self.assertEqual(backend.exec_sql("select count(1) from tc").collect()[0][0], 2)
        pf.set_shuffle_partitions("2")
        self.assertEqual(backend.exec_sql("select count(1) from tc").collect()[0][0], 2)

        iof = IOFuncs(backend.spark)
        iof.write_csv("tc", "file:///tmp/easysql-ut/test_write_csv")
        iof.rename_csv_output("file:///tmp/easysql-ut/test_write_csv", "/tmp/easysql-ut/test_write.csv")
        iof.write_json_local("tc", "/tmp/easysql-ut/test_write.json")
        self.assertEqual(len(json.loads(Path("/tmp/easysql-ut/test_write.json").read_text())), 2)
        with open("/tmp/easysql-ut/test_write.json", "w") as test_write_file:
            json.dump({"a": "0"}, test_write_file)
        iof.update_json_local(
            ProcessorContext(VarsContext({"a": 1}, {"b": [1, 2]}), TemplatesContext()),
            "a,",
            "b",
            "",
            "/tmp/easysql-ut/test_write.json",
        )
        self.assertEqual(json.loads(Path("/tmp/easysql-ut/test_write.json").read_text())["a"], 1)
        self.assertEqual(json.loads(Path("/tmp/easysql-ut/test_write.json").read_text())["b"], [1, 2])
        iof.update_json_local(
            ProcessorContext(VarsContext({"a": 2}, {"b": [1, 2]}), TemplatesContext()),
            "a,",
            "",
            "c.d",
            "/tmp/easysql-ut/test_write.json",
        )
        self.assertEqual(json.loads(Path("/tmp/easysql-ut/test_write.json").read_text())["c"]["d"]["a"], 2)

    def run_test(self, backend: Backend, types: Tuple[str, str, str]):
        try:
            self._run_test(backend, types)
        finally:
            backend.clean()

    def _run_test(self, backend: Backend, types: Tuple[str, str, str]):
        table_name = self.test_table_name

        int_type, str_type, pt_type = types
        if isinstance(backend, RdbBackend):
            backend.exec_native_sql(backend.sql_dialect.drop_db_sql("t"))
        backend.create_table_with_data(
            table_name,
            [["1", 1, "2022-01-01"], ["2", None, "2022-01-02"]],
            [Col("id", str_type), Col("val", int_type), Col("pt", pt_type)],
            [Partition(field="pt")],
        )

        pf = PartitionFuncs(backend) if isinstance(backend, RdbBackend) else SparkPartitionFuncs(backend)

        class _ReportCollector(ReportCollector):
            def collect_report(self, step: Step, status: str = None, message: str = None):
                pass

        step = Step("1", _ReportCollector(), FuncRunner({"bool": lambda x: x == "1"}), select_sql="select 0 as a")

        self.assertEqual(pf.get_first_partition(table_name), "2022-01-01")
        self.assertTrue(pf.is_first_partition(table_name, "2022-01-01"))
        self.assertTrue(pf.is_not_first_partition(table_name, "2022-01-02"))
        self.assertTrue(pf.partition_exists(table_name, "2022-01-02"))
        self.assertFalse(pf.partition_exists(table_name, "2022-01-03"))
        self.assertTrue(pf.partition_not_exists(table_name, "2022-01-03"))
        self.assertTrue(pf.ensure_partition_or_first_partition_exists(step, table_name, "2022-01-01"))
        self.assertTrue(pf.ensure_partition_or_first_partition_exists(step, table_name, "2022-01-02"))
        self.assertTrue(pf.ensure_partition_exists(step, table_name, "2022-01-01"))
        self.assertFalse(pf.previous_partition_exists(table_name, "2022-01-01"))
        self.assertTrue(pf.previous_partition_exists(table_name, "2022-01-03"))
        self.assertFalse(pf.ensure_partition_exists(step, table_name, "2022-01-03"))
        self.assertTrue(pf.ensure_dwd_partition_exists(step, table_name, "2022-01-01", "id"))
        self.assertFalse(pf.ensure_dwd_partition_exists(step, table_name, "2022-01-02", "val"))
        self.assertEqual(pf.get_partition_or_first_partition(table_name, "2021-01-02"), "2022-01-01")
        self.assertEqual(pf.get_partition_or_first_partition(table_name, "2022-01-02"), "2022-01-02")
        self.assertEqual(pf.get_first_partition_optional(table_name), "2022-01-01")
        self.assertEqual(pf.get_last_partition(table_name), "2022-01-02")
        self.assertEqual(pf.get_partition_cols(table_name), ["pt"])
        self.assertEqual(pf.get_partition_col(table_name), "pt")
        self.assertTrue(pf.has_partition_col(table_name))

        f = ColumnFuncs(backend)
        self.assertEqual(f.all_cols_with_exclusion_expr(table_name, "pt"), "func_test.id, func_test.val")
        self.assertEqual(f.all_cols_without_one_expr(table_name, "pt"), "func_test.id, func_test.val")

        f = TableFuncs(backend)

        self.assertFalse(f.ensure_no_null_data_in_table(step, table_name))
        self.assertTrue(f.ensure_no_null_data_in_table(step, table_name, "id='1'"))
        self.assertTrue(f.check_not_null_column_in_table(step, table_name, "id"))
        self.assertFalse(f.check_not_null_column_in_table(step, table_name, "val"))
        self.assertTrue(f.check_not_null_column_in_table(step, table_name, "val", "id='1'"))

        class _Alerter(Alerter):
            def __init__(self):
                self.alert_msg = None

            def send_alert(self, msg: str, mentioned_users: str = ""):
                self.alert_msg = msg

        alerter = _Alerter()
        f = AlertFunc(backend, alerter)
        f.alert(
            step,
            ProcessorContext(VarsContext({"a": 1}), TemplatesContext()),
            "test",
            "bool({a})",
            "result: {a}",
            "a,b,c",
        )
        self.assertTrue(alerter.alert_msg is not None)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Tuple, Union

from ...logger import logger
from ...udf import udfs
from ..common import SqlProcessorAssertionError
from .base import Backend, Col, Partition, Row, SaveMode, Table, TableMeta

__all__ = ["SparkRow", "SparkTable", "SparkBackend"]

if TYPE_CHECKING:
    from pyspark.sql import DataFrame
    from pyspark.sql.types import StructType


class SparkRow(Row):
    def __init__(self, row):
        from pyspark.sql.types import Row

        self.row: Row = row

    def as_dict(self):
        return self.row.asDict()

    def as_tuple(self) -> Tuple:
        return self.row

    def __eq__(self, other):
        return self.row.__eq__(other)

    def __str__(self):
        return str(self.row)[4:-1]

    def __getitem__(self, i):
        return self.row[i]

    def __repr__(self):
        return self.row.__repr__()


class SparkTable(Table):
    def __init__(self, df):
        from pyspark.sql import DataFrame

        self.df: DataFrame = df

    def is_empty(self) -> bool:
        return self.df.rdd.isEmpty()

    def field_names(self) -> List[str]:
        return self.df.schema.fieldNames()

    def first(self) -> Row:
        return SparkRow(self.df.first())

    def limit(self, count: int) -> SparkTable:
        return SparkTable(self.df.limit(count))

    def with_column(self, name: str, value: any) -> SparkTable:
        from pyspark.sql import Column
        from pyspark.sql.functions import expr

        return SparkTable(self.df.withColumn(name, value if isinstance(value, Column) else expr(value)))

    def collect(self) -> List[Row]:
        return [SparkRow(row) for row in self.df.collect()]

    def show(self, count: int = 20):
        self.df.show(count)

    def count(self) -> int:
        return self.df.count()


class SparkBackend(Backend):
    def __init__(self, spark, scala_udf_initializer: str = None):
        from pyspark.sql import SparkSession

        self.spark: SparkSession = spark
        self.scala_udf_initializer = scala_udf_initializer

    def reset(self):
        pass

    def init_udfs(self, scala_udf_initializer: str = None, *args, **kwargs):
        scala_udf_initializer = scala_udf_initializer or self.scala_udf_initializer
        if scala_udf_initializer:
            from py4j.java_gateway import java_import

            gw = self.spark.sparkContext._gateway
            java_import(gw.jvm, scala_udf_initializer)
            initUdfs = eval(f"gw.jvm.{scala_udf_initializer}.initUdfs", {"gw": gw})
            initUdfs(self.spark._jsparkSession)

        self.register_udfs(udfs.get_udfs("spark"))

    def register_udfs(self, funcs: Dict[str, Callable]):
        for key in funcs:
            func = funcs[key]
            self.spark.udf.register(key, func)

    def set_spark_configs(self, configs: Dict[str, str]):
        for key, value in configs.items():
            self.spark.conf.set(key, value)

    def temp_tables(self) -> List[str]:
        return [table.name for table in self.spark.catalog.listTables("default") if table.isTemporary]

    def clear_cache(self):
        self.spark.catalog.clearCache()

    def clear_temp_tables(self, exclude: List[str] = None):
        exclude = exclude or []
        for table in self.spark.catalog.listTables("default"):
            if table.isTemporary and table.name not in exclude:
                print(f"dropping temp view {table.name}")
                self.spark.catalog.dropTempView(table.name)

    def create_empty_table(self):
        from pyspark.sql.types import StructType

        return SparkTable(self.spark.createDataFrame(self.spark.sparkContext.emptyRDD(), StructType([])))

    def create_temp_table(self, table: SparkTable, name: str):
        table.df.createOrReplaceTempView(name)

    def create_cache_table(self, table: SparkTable, name: str):
        table.df.createOrReplaceTempView(name)
        self.spark.catalog.cacheTable(name)

    def broadcast_table(self, table: SparkTable, name: str):
        from pyspark.sql.functions import broadcast

        df = broadcast(table.df)
        df.createOrReplaceTempView(name)

    def exec_native_sql(self, sql: str) -> DataFrame:
        logger.info(f"will exec sql: {sql}")
        return self.spark.sql(sql)

    def exec_sql(self, sql: str) -> Table:
        logger.info(f"will exec sql: {sql}")
        return SparkTable(self.spark.sql(sql))

    def table_exists(self, table: TableMeta):
        from pyspark.sql.utils import AnalysisException

        try:
            return self.spark._jsparkSession.catalog().tableExists(table.dbname, table.pure_table_name)
        except AnalysisException:
            return False

    def _create_table(self, dbname: str, table_name: str, schema: StructType, partitions: List[Partition]):
        from pyspark.sql.functions import lit

        spark = self.spark
        df = spark.createDataFrame(spark.sparkContext.emptyRDD(), schema)
        schema = df.schema
        for p in partitions or []:
            if p.field not in schema.fieldNames():
                if p.value is None:
                    raise SqlProcessorAssertionError(
                        "partition column value is None when create table with partitions but partitions is not in"
                        f" dataframe. this should not happen. table_name={dbname}.{table_name}, p.field={p.field},"
                        f" p.value={p.value}"
                    )
                df = df.withColumn(p.field, lit(p.value))

        df.createOrReplaceTempView("table_data")
        partition_expr = f'partitioned by ({",".join([p.field for p in partitions])}) ' if partitions else ""
        create_database_stmt = f"create database if not exists {dbname}"
        create_table_stmt = f"""create table if not exists {dbname}.{table_name} using hive
                                options(FILEFORMAT "parquet") {partition_expr}
                                TBLPROPERTIES ("transactional" = "false")
                                as select * from table_data"""

        self.exec_native_sql(create_database_stmt)
        self.exec_native_sql(create_table_stmt)
        return self

    def save_table(
        self,
        source_table_meta: TableMeta,
        target_table_meta: TableMeta,
        save_mode: SaveMode,
        create_target_table: bool,
    ):
        from pyspark.sql.functions import lit

        if not self.table_exists(target_table_meta) and create_target_table:
            schema = self.spark.sql(f"select * from {source_table_meta.table_name}").limit(0).schema
            self._create_table(
                target_table_meta.dbname, target_table_meta.pure_table_name, schema, target_table_meta.partitions
            )

        temp_res = self.exec_native_sql(f"select * from {source_table_meta.table_name}")
        # partial dynamic partition (动态分区和静态分区同时使用）在 spark 2.3.2 上有问题，参见 https://issues.apache.org/jira/browse/SPARK-31605
        # 纯动态分区时，如果当日没有新增数据，则不会创建 partition。而我们希望对于静态分区，总是应该创建分区，即使当日没有数据
        dynamic_partitions = list(filter(lambda p: not p.value, target_table_meta.partitions))
        static_partitions = list(filter(lambda p: p.value, target_table_meta.partitions))
        columns = self.exec_native_sql(f"select * from {target_table_meta.table_name}").limit(0).columns
        if dynamic_partitions:
            for p in static_partitions:
                temp_res = temp_res.withColumn(p.field, lit(p.value))
            temp_res = temp_res.select(*columns)
            fields = [f"{p.field}" for p in dynamic_partitions]
        else:
            columns = [c for c in columns if c not in [p.field for p in static_partitions]]
            temp_res = temp_res.select(*columns)
            fields = [
                f"{p.field}='{p.value}'" if isinstance(p.value, str) else f"{p.field}={p.value}"
                for p in target_table_meta.partitions
            ]
        partition_expr = f"partition ({','.join(fields)})" if fields else ""

        # to resolve issue: pyspark.sql.utils.AnalysisException: Cannot overwrite a path that is also being read from.
        # refer: https://stackoverflow.com/questions/38746773/read-from-a-hive-table-and-write-back-to-it-using-spark-sql
        temp_res = self.spark.createDataFrame(temp_res.rdd, temp_res.schema)
        temp_res.createOrReplaceTempView("res")

        save_sql = (
            f"insert {'into' if save_mode == SaveMode.append else save_mode.name} table"
            f" {target_table_meta.table_name} {partition_expr} select * from res"
        )
        self.exec_native_sql(save_sql)

    def refresh_table_partitions(self, table: TableMeta):
        df = self.exec_native_sql(f"desc {table.table_name}")
        column_list = df.select(df.col_name, df.data_type).rdd.map(lambda x: (x[0], x[1])).collect()
        partition_details = [
            column_list[index + 1 :] for index, item in enumerate(column_list) if item[0] == "# col_name"
        ]
        if len(partition_details) == 0:
            table.update_partitions([])
        else:
            partitions = [Partition(x[0]) for x in partition_details[0]]
            table.update_partitions([Partition(p.field, p.value) for p in partitions])

    def clean(self):
        from easy_sql.spark_optimizer import clear_temp_views

        self.spark.catalog.clearCache()
        clear_temp_views(self.spark)

    def create_table_with_data(
        self,
        full_table_name: str,
        values: List[List[Any]],
        schema: Union[StructType, List[Col]],
        partitions: List[Partition],
    ):
        print(f"creating table: {full_table_name}")
        self.spark.sql(f'create database if not exists {full_table_name.split(".")[0]}')
        self.spark.sql(f"drop table if exists {full_table_name}").collect()
        from pyspark.sql.types import StructType

        schema_or_cols = schema if isinstance(schema, StructType) else [col.name for col in schema]
        write = self.spark.createDataFrame(values, schema_or_cols).write
        if partitions:
            write = write.partitionBy(*[p.field for p in partitions])
        write.mode("overwrite").saveAsTable(full_table_name, mode="overwrite")

    def create_temp_table_with_data(self, table_name: str, values: List[List[Any]], schema: StructType):
        self.spark.createDataFrame(values, schema).createOrReplaceTempView(table_name)



import unittest

from easy_sql.sql_linter.sql_linter import SqlLinter


class SqlLinterTest(unittest.TestCase):
    def test_should_work_for_normal_sql(self):
        sql = """select  a,
date,date2,
aa from sales_order
"""

        sql_linter = SqlLinter(sql, include_rules=None, exclude_rules=None)
        result = sql_linter.lint("bigquery", easysql=False)
        fixed = sql_linter.fix("bigquery", easy_sql=False)
        self.assertEqual(len(result), 7)
        print(fixed)
        expected = """select a
    , date, date2
    , aa from ${temp_db}.sales_order
"""
        self.assertEqual(expected, fixed)

    def test_should_work_for_exclude_rules(self):
        sql = """select  a,
date,date2,
aa from sales_order
"""

        sql_linter = SqlLinter(sql, include_rules=None, exclude_rules=["BigQuery_L001"])
        result = sql_linter.lint("bigquery", easysql=False)
        fixed = sql_linter.fix("bigquery", easy_sql=False)
        self.assertEqual(len(result), 6)
        print(fixed)
        expected = """select a
    , date, date2
    , aa from sales_order
"""
        self.assertEqual(expected, fixed)

    def test_should_work_for_variables(self):
        sql = """-- backend: bigquery
-- target=temp.feature_stage_0_out
select  a,
${date},${date2},
${${aa}} from sales_order
"""

        sql_linter = SqlLinter(sql, include_rules=None, exclude_rules=None)
        result = sql_linter.lint("bigquery")
        fixed = sql_linter.fix("bigquery")
        self.assertEqual(len(result), 7)
        expected = """-- backend: bigquery
-- target=temp.feature_stage_0_out
select a
    , ${date}, ${date2}
    , ${${aa}} from ${temp_db}.sales_order
"""
        self.assertEqual(expected, fixed)

    def test_should_work_when_have_three_quote(self):
        sql = '''-- backend: bigquery
-- target=temp.feature_stage_0_out
select  a,"""hh"""ab, "" ac from table
'''

        sql_linter = SqlLinter(sql, include_rules=None, exclude_rules=None)
        result = sql_linter.lint("bigquery")
        fixed = sql_linter.fix("bigquery")
        self.assertEqual(len(result), 5)
        expected = '''-- backend: bigquery
-- target=temp.feature_stage_0_out
select a, """hh""" as ab, "" as ac from ${temp_db}.table
'''
        self.assertEqual(expected, fixed)

    def test_should_work_for_multiple_step(self):
        sql = """-- backend: bigquery
-- target=variables
select
    translate('${data_date}', '-', '')  as data_date_no_ds
    , true                              as __create_hive_table__
    
-- target=temp.model_data
select * from sales_model_demo_with_label where date='${data_date_no_ds}'


-- target=temp.feature_stage_0_out
select *
, ${data_date_no_ds}, a1, b1 from ${temp_db}.model_data
"""

        sql_linter = SqlLinter(sql, include_rules=None, exclude_rules=None)
        result = sql_linter.lint("bigquery")
        self.assertEqual(len(result), 7)
        fixed = sql_linter.fix("bigquery")
        expected = """-- backend: bigquery
-- target=variables
select
    translate('${data_date}', '-', '') as data_date_no_ds
    , true as __create_hive_table__

-- target=temp.model_data
select
    *
from ${temp_db}.sales_model_demo_with_label where date = '${data_date_no_ds}'

-- target=temp.feature_stage_0_out
select *
    , ${data_date_no_ds}, a1, b1 from ${temp_db}.model_data
"""
        self.assertEqual(fixed, expected)

    def test_should_work_when_have_template(self):
        # bigquery union must be union all
        sql = """-- backend: bigquery
-- target=template.dim_cols
product_name
, product_category

-- target=temp.dims
select @{dim_cols} from order_count
union all
select @{dim_cols} from sales_amount

-- target=temp.result
@{transited_with_fk_inc_template(
    source_table_name=transited, source_table_biz_key=product_id, result_col_name=product_id_key,
    fk_table_name=dwd_sales.sales_dim_product_h, pk=product_key, fk_table_biz_key=id,
    update_time_col_name=update_time, partition_col_name=di)} 
"""
        sql_linter = SqlLinter(sql, exclude_rules=["L025"])
        result = sql_linter.lint("bigquery")
        self.assertEqual(len(result), 2)
        fixed = sql_linter.fix("bigquery")
        print(fixed)
        expected = """-- backend: bigquery
-- target=template.dim_cols
product_name
, product_category

-- target=temp.dims
select @{dim_cols} from ${temp_db}.order_count
union all
select @{dim_cols} from ${temp_db}.sales_amount

-- target=temp.result
@{transited_with_fk_inc_template(
    source_table_name=transited, source_table_biz_key=product_id, result_col_name=product_id_key,
    fk_table_name=dwd_sales.sales_dim_product_h, pk=product_key, fk_table_biz_key=id,
    update_time_col_name=update_time, partition_col_name=di)} """
        self.assertEqual(expected, fixed)

    def test_should_work_when_given_diff_backend(self):
        # spark can only use union
        sql = """-- backend: spark
-- --------------------
-- 标签：产品销量
-- --------------------

-- config: spark.master=local[2]
-- config: spark.submit.deployMode=client
-- config: spark.executor.memory=1g
-- config: spark.executor.cores=1

-- inputs: dwd_sales.sales_fact_order_h, dwd_sales.sales_dim_product_h
-- outputs: dm_sales.product_label_amount
-- owner: all

-- target=template.dim_cols
product_name
, product_category

-- target=temp.dims
select @{dim_cols} from order_count
union 
select @{dim_cols} from sales_amount
"""
        sql_linter = SqlLinter(sql)
        result = sql_linter.lint("spark", True)
        self.assertEqual(len(result), 1)
        fixed = sql_linter.fix("spark")
        expected = """-- backend: spark
-- --------------------
-- 标签：产品销量
-- --------------------

-- config: spark.master=local[2]
-- config: spark.submit.deployMode=client
-- config: spark.executor.memory=1g
-- config: spark.executor.cores=1

-- inputs: dwd_sales.sales_fact_order_h, dwd_sales.sales_dim_product_h
-- outputs: dm_sales.product_label_amount
-- owner: all

-- target=template.dim_cols
product_name
, product_category

-- target=temp.dims
select @{dim_cols} from order_count
union
select @{dim_cols} from sales_amount
"""
        self.assertEqual(fixed, expected)

    def test_should_work_when_have_functions(self):
        sql = """-- backend: bigquery

-- target=variables
select ${plus(1, 2)} as a
, flag as b
from data_table 

-- target=temp.dims
select * from order_count where value < ${a}

-- target=func.plus(1, 1)

"""
        sql_linter = SqlLinter(sql)
        result = sql_linter.lint("bigquery", True)
        print("after fix")
        print(len(result))
        fixed = sql_linter.fix("bigquery")
        print(fixed)
        self.assertEqual(len(result), 4)
        expected = """-- backend: bigquery

-- target=variables
select ${plus(1, 2)} as a
    , flag as b
from ${temp_db}.data_table

-- target=temp.dims
select * from ${temp_db}.order_count where value < ${a}

-- target=func.plus(1, 1)"""
        self.assertEqual(expected, fixed)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections import Callable

    from pyspark.sql.types import DataType


class UDF:
    def __init__(self, func: Callable, return_type: DataType):
        self.func = func
        self.return_type = return_type

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


def check_regex_func(pattern):
    return lambda any_str: any_str if any_str and re.match(pattern, any_str) else None



import json
import os
from typing import List

from .backend.rdb import RdbBackend
from .common import is_int_type
from .funcs_common import AlertFunc, ColumnFuncs
from .funcs_common import PartitionFuncs as PartitionFuncsBase
from .funcs_common import TableFuncs

__all__ = ["PartitionFuncs", "ColumnFuncs", "AlertFunc", "TableFuncs", "ModelFuncs"]


class ModelFuncs:
    def __init__(self, backend: RdbBackend):
        self.backend = backend

    def bq_model_predict_with_local_spark(
        self,
        model_save_path: str,
        input_table_name: str,
        output_table_name: str,
        feature_cols: str,
        id_col: str,
        output_ref_cols: str,
    ):
        from pyspark.ml import PipelineModel
        from pyspark.sql.functions import expr

        from ..spark_optimizer import get_spark

        SPARK_JARS = (
            "/tmp/app/dataplat/lib/scala/lib_spark3/spark-bigquery-latest_2.12.jar,"
            "/tmp/app/dataplat/lib/scala/lib_spark3/gcs-connector-hadoop2-latest.jar,"
        )

        spark_config_settings_dict = {
            "spark.jars": SPARK_JARS,
            "spark.master": "local[2]",
            "spark.submit.deployMode": "client",
            "spark.executor.memory": "1g",
            "spark.executor.cores": "1",
            "spark.hadoop.fs.AbstractFileSystem.gs.impl": "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFS",
            "parentProject": "data-dev-workbench-prod-3fd9",
            "spark.sql.warehouse.dir": "/tmp/spark-warehouse-localdw",
            "spark.driver.extraJavaOptions": (
                "-Dderby.system.home=/tmp/spark-warehouse-metastore "
                "-Dderby.stream.error.file=/tmp/spark-warehouse-metastore.log"
            ),
            "credentialsFile": f"{os.environ.get('HOME', '/tmp')}/.bigquery/credential-prod.json",
        }

        spark = get_spark("test", spark_config_settings_dict)
        bucket = "dataplat-gcp-demo"
        spark.conf.set("temporaryGcsBucket", bucket)

        spark._jsc.hadoopConfiguration().set("fs.gs.impl", "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFileSystem")
        spark._jsc.hadoopConfiguration().set("fs.gs.auth.service.account.enable", "true")
        spark._jsc.hadoopConfiguration().set(
            "google.cloud.auth.service.account.json.keyfile",
            f"{os.environ.get('HOME', '/tmp')}/.bigquery/credential-prod.json",
        )

        spark.read.format("bigquery").option("parentProject", "data-dev-workbench-prod-3fd9").option(
            "table", input_table_name
        ).load().createOrReplaceTempView("origin_input")

        data = spark.sql(f"select {feature_cols} from origin_input")

        output_ref_cols = [col.strip() for col in output_ref_cols.split(",") if col.strip()]
        model = PipelineModel.load(model_save_path)

        int_cols = [f.name for f in data.schema.fields if is_int_type(f.dataType.typeName())]
        for col in int_cols:
            data = data.withColumn(col, expr(f"cast({col} as double)"))

        predictions = model.transform(data)
        output = predictions.select(output_ref_cols + [id_col, "prediction"])

        output.write.format("bigquery").option("parentProject", "data-dev-workbench-prod-3fd9").option(
            "table", output_table_name
        ).mode("overwrite").save()

    def model_predict_with_local_spark(
        self,
        model_save_path: str,
        input_table_name: str,
        output_table_name: str,
        feature_cols: str,
        id_col: str,
        output_ref_cols: str,
    ):
        from pyspark.ml import PipelineModel
        from pyspark.sql.functions import expr

        from ..spark_optimizer import get_spark

        workdir = os.path.dirname(os.environ.get("PWD"))
        local_settings_dir = f"{workdir}/settings.local.json"
        settings_dir = f"{workdir}/settings.json"

        with open(settings_dir, "r") as json_file:
            settings = json.load(json_file)
            data_source_load_options = settings["ml_model"]["data_source_load_options"]
            data_type_map = settings["ml_model"]["data_type_map"]
            SPARK_JARS = "".join(settings["ml_model"]["spark_jars"])

        with open(local_settings_dir, "r") as json_file:
            local_settings = json.load(json_file)
            local_data_source_load_options = local_settings["ml_model"]["data_source_load_options"]

        file_dir = os.path.dirname(os.path.abspath(__file__))

        spark_config_settings_dict = {
            "spark.jars": SPARK_JARS,
            "spark.master": "local[2]",
            "spark.submit.deployMode": "client",
            "spark.executor.memory": "1g",
            "spark.executor.cores": "1",
            "spark.sql.warehouse.dir": "/tmp/spark-warehouse-localdw",
            "spark.driver.extraJavaOptions": (
                "-Dderby.system.home=/tmp/spark-warehouse-metastore "
                "-Dderby.stream.error.file=/tmp/spark-warehouse-metastore.log"
            ),
            "spark.driver.extraClassPath": os.path.join(file_dir, "../../../deps/lib/*"),
        }

        from py4j.java_gateway import java_import

        spark = get_spark("ml_local", spark_config_settings_dict)
        gw = spark.sparkContext._gateway
        java_import(gw.jvm, "common.dataplat.sparkutils")

        data = (
            spark.read.format(f"{data_source_load_options['format']}")
            .option("driver", f"{data_source_load_options['driver']}")
            .option("url", f"{data_source_load_options['url']}")
            .option("user", f"{data_source_load_options['user']}")
            .option("password", f"{local_data_source_load_options['password']}")
            .option("dbtable", input_table_name)
            .load()
        )

        output_ref_cols = [col.strip() for col in output_ref_cols.split(",") if col.strip()]
        model = PipelineModel.load(model_save_path)

        int_cols = [f.name for f in data.schema.fields if is_int_type(f.dataType.typeName())]
        for col in int_cols:
            data = data.withColumn(col, expr(f"cast({col} as double)"))

        predictions = model.transform(data)
        output = predictions.select(output_ref_cols + [id_col, "prediction"])

        ddl = self.__get_ddl_by_backend(data_type_map, output_table_name, id_col, output.dtypes)

        self.backend.exec_native_sql(ddl)
        self.backend.exec_native_sql(f"truncate table {output_table_name}")
        output.write.format(f"{data_source_load_options['format']}").mode("append").option(
            "driver", f"{data_source_load_options['driver']}"
        ).option("truncate", f"{data_source_load_options['truncate']}").option(
            "url", f"{data_source_load_options['url']}"
        ).option(
            "user", f"{data_source_load_options['user']}"
        ).option(
            "password", f"{local_data_source_load_options['password']}"
        ).option(
            "dbtable", output_table_name
        ).save()

    def __get_ddl_by_backend(self, data_type_map, output_table_name, id_col, output_dtypes) -> str:
        if self.backend.is_postgres_backend:
            columns_def = ", ".join(
                [
                    f'{col} {data_type_map[col_type]} {"PRIMARY KEY" if col == id_col else ""}'
                    for col, col_type in output_dtypes
                ]
            )
            return f"CREATE TABLE IF NOT EXISTS {output_table_name} ({columns_def})"
        elif self.backend.is_clickhouse_backend:
            columns_def = ", ".join([f"{col} {data_type_map[col_type]}" for col, col_type in output_dtypes])
            return f"""CREATE TABLE IF NOT EXISTS {output_table_name} ({columns_def}) engine = MergeTree
            PRIMARY KEY {id_col}
            ORDER BY {id_col}"""
        else:
            msg = (
                "Backend of type"
                f' {type(self.backend)}-{self.backend.backend_type if isinstance(self.backend, RdbBackend) else ""} is'
                " not supported yet"
            )
            raise Exception(msg)


class PartitionFuncs(PartitionFuncsBase):
    def __check_backend(self):
        if not isinstance(self.backend, RdbBackend):
            msg = (
                "Backend of type"
                f' {type(self.backend)}-{self.backend.backend_type if isinstance(self.backend, RdbBackend) else ""} is'
                " not supported yet"
            )
            raise Exception(msg)

    def _get_bigquery_partition_values(self, table_name):
        db, table = self.__parse_table_name(table_name)
        sql = (
            f"select distinct partition_value from {db}.__table_partitions__ where table_name = '{table}' order by"
            " partition_value"
        )
        partition_values = [str(v[0]) for v in self.backend.exec_sql(sql).collect()]
        return partition_values

    def _get_clickhouse_partition_values(self, table_name):
        db, table = self.__parse_table_name(table_name)
        sql = (
            f"SELECT distinct partition_value FROM {self.backend.partitions_table_name} where db_name = '{db}' and"
            f" table_name = '{table}';"
        )
        partition_values = [str(v[0]) for v in self.backend.exec_sql(sql).collect()]
        partition_values.sort()
        return partition_values

    def _get_postgresql_partition_values(self, table_name):
        db, table = self.__parse_table_name(table_name)
        sql = f"""
        SELECT
            concat(nmsp_child.nspname, '.', child.relname) as partition_tables,
            pg_catalog.pg_get_expr(child.relpartbound, child.oid) as partition_expr
        FROM pg_inherits
            JOIN pg_class parent        ON pg_inherits.inhparent = parent.oid
            JOIN pg_class child         ON pg_inherits.inhrelid   = child.oid
            JOIN pg_namespace nmsp_parent   ON nmsp_parent.oid  = parent.relnamespace
            JOIN pg_namespace nmsp_child    ON nmsp_child.oid   = child.relnamespace
            JOIN pg_partitioned_table part  ON part.partrelid = parent.oid
        WHERE nmsp_parent.nspname='{db}' and parent.relname='{table}'
        """
        partition_values = [str(v[1]) for v in self.backend.exec_sql(sql).collect()]
        for p in partition_values:
            if not p.upper().startswith("FOR VALUES FROM (") or ") TO (" not in p.upper():
                raise Exception("unable to parse partition: " + p)
        partition_values = [v[len("FOR VALUES FROM (") : v.upper().index(") TO (")] for v in partition_values]
        partition_values = [v.strip("'") if v.startswith("'") else int(v) for v in partition_values]
        partition_values.sort()
        return partition_values

    def _get_partition_values(self, table_name):
        self.__check_backend()
        if self.backend.is_pg:
            return self._get_postgresql_partition_values(table_name)
        elif self.backend.is_ch:
            return self._get_clickhouse_partition_values(table_name)
        elif self.backend.is_bq:
            return self._get_bigquery_partition_values(table_name)
        else:
            msg = (
                "Backend of type"
                f' {type(self.backend)}-{self.backend.backend_type if isinstance(self.backend, RdbBackend) else ""} is'
                " not supported yet"
            )
            raise Exception(msg)

    def get_partition_cols(self, table_name: str) -> List[str]:
        self.__check_backend()
        db, table = self.__parse_table_name(table_name)
        if isinstance(self.backend, RdbBackend):
            native_partitions_sql, extract_partition_cols = self.backend.sql_dialect.native_partitions_sql(
                f"{db}.{table}"
            )
            pt_cols = extract_partition_cols(self.backend.exec_native_sql(native_partitions_sql))
            return pt_cols
        else:
            raise AssertionError("should not happen!")

    def __parse_table_name(self, table_name):
        backend: RdbBackend = self.backend
        full_table_name = table_name if "." in table_name else f"{backend.temp_schema}.{table_name}"
        db, table = full_table_name[: full_table_name.index(".")], full_table_name[full_table_name.index(".") + 1 :]
        return db, table



from typing import Dict


def get_attr(obj: Dict, path: str):
    data_current = obj
    if not path:
        return data_current
    for attr_current in path.split("."):
        assert attr_current != "", f"Neither part of path should be empty: path=`{path}`, current_part=`{attr_current}`"
        if attr_current not in data_current:
            data_current[attr_current] = {}
        data_current = data_current[attr_current]
    return data_current



import re
import unittest
from functools import partial

from easy_sql import base_test
from easy_sql.base_test import dt
from easy_sql.sql_processor.backend import Partition, SaveMode, TableMeta
from easy_sql.sql_processor.backend.maxcompute import MaxComputeBackend, MaxComputeRow


class MaxComputeTest(unittest.TestCase):
    def test_mc_table(self):
        if not base_test.should_run_integration_test("mc"):
            return

        # TODO: fetch the account info from environment
        odps_parms = {
            "access_id": "xx",
            "secret_access_key": "xx",
            "project": "xx",
            "endpoint": "http://service.cn-hangzhou.maxcompute.aliyun.com/api",
        }
        backend = MaxComputeBackend(**odps_parms)

        temp_table_name = "test_mc_unit_test"
        backend.conn.execute_sql(f"drop table if exists {temp_table_name}")
        backend.conn.execute_sql(f"create table {temp_table_name}(id int, val varchar(1000))")
        backend.conn.execute_sql(f"insert into {temp_table_name} values(1, '1'), (2, '2'), (3, '3')")

        self.run_test_table(backend, temp_table_name)

        backend.conn.execute_sql(f"drop table if exists {temp_table_name}")
        backend.clear_temp_tables()

    def run_test_table(self, backend: MaxComputeBackend, table_name):
        table = backend.exec_sql(f"select * from {table_name}")
        self.assertFalse(table.is_empty())

        table = table.limit(0)
        self.assertTrue(table.is_empty())

        table = backend.exec_sql(f"select * from {table_name} order by id").with_column("a", 1)
        self.assertFalse(table.is_empty())
        self.assertTrue(table.count(), 3)
        cols, types = ["id", "val", "a"], ["int", "varchar(1000)", "int"]
        self.assertListEqual(
            table.collect(),
            [
                MaxComputeRow.from_schema_meta(cols, types, (1, "1", 1)),
                MaxComputeRow.from_schema_meta(cols, types, (2, "2", 1)),
                MaxComputeRow.from_schema_meta(cols, types, (3, "3", 1)),
            ],
        )
        self.assertEqual(table.first(), MaxComputeRow.from_schema_meta(cols, types, (1, "1", 1)))
        self.assertListEqual(table.field_names(), cols)
        self.assertEqual(table.limit(2).count(), 2)

        table = backend.exec_sql(f"select * from {table_name} order by id").with_column("a", "1")
        cols, types = ["id", "val", "a"], ["int", "varchar(1000)", "string"]
        self.assertEqual(table.first(), MaxComputeRow.from_schema_meta(cols, types, (1, "1", "1")))

        table = backend.exec_sql(f"select * from {table_name} order by id").with_column("a", 1.1)
        cols, types = ["id", "val", "a"], ["int", "varchar(1000)", "float"]
        self.assertEqual(table.first(), MaxComputeRow.from_schema_meta(cols, types, (1, "1", 1.1)))

        table = backend.exec_sql(f"select * from {table_name} order by id").with_column("a", dt("2020-01-01 11:11:11"))
        cols, types = ["id", "val", "a"], ["int", "varchar(1000)", "timestamp"]
        self.assertEqual(
            table.first(), MaxComputeRow.from_schema_meta(cols, types, (1, "1", dt("2020-01-01 11:11:11")))
        )

        table.show(1)

    def test_mc_backend(self):
        if not base_test.should_run_integration_test("mc"):
            return

        # TODO: fetch the account info from environment
        odps_parms = {
            "access_id": "xx",
            "secret_access_key": "xx",
            "project": "xx",
            "endpoint": "http://service.cn-hangzhou.maxcompute.aliyun.com/api",
        }
        backend = MaxComputeBackend(**odps_parms)

        temp_table_name = "test_mc_unit_test"
        backend.conn.execute_sql(f"drop table if exists {temp_table_name}")
        backend.conn.execute_sql(f"create table {temp_table_name}(id int, val varchar(1000))")
        backend.conn.execute_sql(f"insert into {temp_table_name} values(1, '1'), (2, '2'), (3, '3')")

        self.run_test_backend(backend, temp_table_name)

        backend.conn.execute_sql(f"drop table if exists {temp_table_name}")
        backend.clear_temp_tables()

    def run_test_backend(self, backend: MaxComputeBackend, table_name):
        backend.create_empty_table()  # should not raise exception

        table = backend.exec_sql(f"select * from {table_name} limit 2")
        backend.create_temp_table(table, "test_limit")
        sc = table.schema
        self.assertListEqual(
            backend.exec_sql("select * from test_limit order by id").collect(),
            [MaxComputeRow(schema=sc, values=(1, "1")), MaxComputeRow(schema=sc, values=(2, "2"))],
        )
        self.assertTrue(backend.table_exists(TableMeta(table_name)))
        self.assertFalse(backend.table_exists(TableMeta(f"{table_name}_xxx")))
        self.assertTrue(backend.table_exists(TableMeta("test_limit")))

        # save without transformation or partitions, should create
        backend.exec_native_sql("drop table if exists xx0")
        backend.create_cache_table(backend.exec_sql(f"select * from {table_name}"), "cache_test")
        backend.save_table(TableMeta("cache_test"), TableMeta("xx0"), SaveMode.overwrite, True)
        self.assertListEqual(
            backend.exec_sql("select * from xx0 order by id").collect(),
            [
                MaxComputeRow(schema=sc, values=(1, "1")),
                MaxComputeRow(schema=sc, values=(2, "2")),
                MaxComputeRow(schema=sc, values=(3, "3")),
            ],
        )
        backend.exec_native_sql("drop table if exists xx0")

        backend.exec_native_sql("drop table if exists xx")
        self.assertRaisesRegex(
            Exception,
            re.compile(r".* cannot save table.*"),
            lambda: backend.save_table(TableMeta("test_limit"), TableMeta("xx"), SaveMode.overwrite, False),
        )
        backend.exec_native_sql("drop table if exists xx")

        # first save without partitions, should create
        backend.save_table(TableMeta("test_limit"), TableMeta("xx"), SaveMode.overwrite, True)
        self.assertListEqual(
            backend.exec_sql("select * from xx order by id").collect(),
            [MaxComputeRow(schema=sc, values=(1, "1")), MaxComputeRow(schema=sc, values=(2, "2"))],
        )

        # second save without partitions, should overwrite
        backend.exec_native_sql("drop view if exists test_limit")
        backend.create_cache_table(backend.exec_sql(f"select * from {table_name} limit 2"), "test_limit")
        backend.save_table(TableMeta("test_limit"), TableMeta("xx"), SaveMode.overwrite, True)
        self.assertListEqual(
            backend.exec_sql("select * from xx order by id").collect(),
            [MaxComputeRow(schema=sc, values=(1, "1")), MaxComputeRow(schema=sc, values=(2, "2"))],
        )

        backend.exec_native_sql("drop view if exists test_limit")
        backend.create_cache_table(backend.exec_sql(f"select * from {table_name} limit 2"), "test_limit")
        with self.assertRaises(Exception):  # noqa: B017
            backend.save_table(
                TableMeta("test_limit"),
                TableMeta("xx", partitions=[Partition("a", dt("2021-01-01 00:00:00"))]),
                SaveMode.overwrite,
                True,
            )
        backend.exec_native_sql("drop table if exists xx")

        # first save with partitions, should create
        backend.exec_native_sql("drop view if exists test_limit")
        backend.create_cache_table(backend.exec_sql(f"select * from {table_name} limit 2"), "test_limit")
        backend.save_table(
            TableMeta("test_limit"),
            TableMeta("xx", partitions=[Partition("a", dt("2021-01-01 00:00:00"))]),
            SaveMode.overwrite,
            True,
        )
        mc_row = partial(
            MaxComputeRow.from_schema_meta,
            cols=["id", "val"],
            types=["int", "string"],
            pt_cols=["a"],
            pt_types=["string"],
        )
        self.assertListEqual(
            backend.exec_sql("select * from xx order by id").collect(),
            [
                mc_row(values=(1, "1", dt("2021-01-01 00:00:00"))),
                mc_row(values=(2, "2", dt("2021-01-01 00:00:00"))),
            ],
        )

        # second save with partitions, should overwrite
        backend.exec_native_sql("drop view if exists test_limit")
        backend.create_cache_table(backend.exec_sql(f"select * from {table_name} limit 2"), "test_limit")
        backend.save_table(
            TableMeta("test_limit"),
            TableMeta("xx", partitions=[Partition("a", dt("2021-01-02 00:00:00"))]),
            SaveMode.overwrite,
            True,
        )
        self.assertListEqual(
            backend.exec_sql("select * from xx order by a, id").collect(),
            [
                mc_row(values=(1, "1", dt("2021-01-01 00:00:00"))),
                mc_row(values=(2, "2", dt("2021-01-01 00:00:00"))),
                mc_row(values=(1, "1", dt("2021-01-02 00:00:00"))),
                mc_row(values=(2, "2", dt("2021-01-02 00:00:00"))),
            ],
        )

        # third save with partitions, should overwrite
        backend.exec_native_sql("drop view if exists test_limit")
        backend.create_cache_table(backend.exec_sql(f"select * from {table_name} limit 2"), "test_limit")
        backend.save_table(
            TableMeta("test_limit"),
            TableMeta("xx", partitions=[Partition("a", dt("2021-01-02 00:00:00"))]),
            SaveMode.overwrite,
            False,
        )
        self.assertListEqual(
            backend.exec_sql("select * from xx order by a, id").collect(),
            [
                mc_row(values=(1, "1", dt("2021-01-01 00:00:00"))),
                mc_row(values=(2, "2", dt("2021-01-01 00:00:00"))),
                mc_row(values=(1, "1", dt("2021-01-02 00:00:00"))),
                mc_row(values=(2, "2", dt("2021-01-02 00:00:00"))),
            ],
        )

        # fourth save with partitions, should append
        backend.exec_native_sql("drop view if exists test_limit")
        backend.create_cache_table(backend.exec_sql(f"select * from {table_name} limit 2"), "test_limit")
        backend.save_table(
            TableMeta("test_limit"),
            TableMeta("xx", partitions=[Partition("a", dt("2021-01-02 00:00:00"))]),
            SaveMode.append,
            False,
        )
        self.assertListEqual(
            backend.exec_sql("select * from xx order by a, id").collect(),
            [
                mc_row(values=(1, "1", dt("2021-01-01 00:00:00"))),
                mc_row(values=(2, "2", dt("2021-01-01 00:00:00"))),
                mc_row(values=(1, "1", dt("2021-01-02 00:00:00"))),
                mc_row(values=(1, "1", dt("2021-01-02 00:00:00"))),
                mc_row(values=(2, "2", dt("2021-01-02 00:00:00"))),
                mc_row(values=(2, "2", dt("2021-01-02 00:00:00"))),
            ],
        )

        backend.exec_native_sql("drop table if exists xx")



import unittest

from easy_sql.base_test import LocalSpark
from easy_sql.sql_processor import SqlProcessor


class FunctionsTest(unittest.TestCase):
    def test_remove_all_whitespaces(self):
        spark = LocalSpark.get()
        SqlProcessor(spark, "")

        self.assertEqual("ab", spark.sql("select remove_all_whitespaces('  a    b  ')").collect()[0][0])
        self.assertEqual("ab", spark.sql("select remove_all_whitespaces(' \ta\t    b\t ')").collect()[0][0])
        self.assertEqual("ab", spark.sql("select remove_all_whitespaces(' \na\n    b\n ')").collect()[0][0])
        self.assertEqual("ab", spark.sql("select remove_all_whitespaces(' \fa\f    b\f ')").collect()[0][0])
        self.assertEqual("ab", spark.sql("select remove_all_whitespaces(' \ra\r    b\r ')").collect()[0][0])
        self.assertEqual("ab", spark.sql("select remove_all_whitespaces(' \va\v    b\v ')").collect()[0][0])

        self.assertEqual("ab", spark.sql("select remove_all_whitespaces('  a     b  ')").collect()[0][0])  # \u00A0
        self.assertEqual("ab", spark.sql("select remove_all_whitespaces('  a     b  ')").collect()[0][0])  # \u2007
        self.assertEqual("ab", spark.sql("select remove_all_whitespaces('  a     b  ')").collect()[0][0])  # 202F
        self.assertEqual(None, spark.sql("select remove_all_whitespaces(NULL)").collect()[0][0])
        self.assertEqual("", spark.sql("select remove_all_whitespaces('')").collect()[0][0])
        self.assertEqual("", spark.sql("select remove_all_whitespaces('  ')").collect()[0][0])

    def test_trim_all(self):
        spark = LocalSpark.get()
        SqlProcessor(spark, "")

        self.assertEqual("a    b", spark.sql("select trim_all('  a    b  ')").collect()[0][0])
        self.assertEqual("a    b", spark.sql("select trim_all(' \ta    b\t ')").collect()[0][0])
        self.assertEqual("a    b", spark.sql("select trim_all(' \na    b\n ')").collect()[0][0])
        self.assertEqual("a    b", spark.sql("select trim_all(' \fa    b\f ')").collect()[0][0])
        self.assertEqual("a    b", spark.sql("select trim_all(' \ra    b\r ')").collect()[0][0])
        self.assertEqual("a    b", spark.sql("select trim_all(' \va    b\v ')").collect()[0][0])

        self.assertEqual("a    b", spark.sql("select trim_all('  a    b  ')").collect()[0][0])  # \u00A0
        self.assertEqual("a    b", spark.sql("select trim_all('  a    b  ')").collect()[0][0])  # \u2007
        self.assertEqual("a    b", spark.sql("select trim_all('  a    b  ')").collect()[0][0])  # 202F
        self.assertEqual(None, spark.sql("select trim_all(NULL)").collect()[0][0])
        self.assertEqual("", spark.sql("select trim_all('')").collect()[0][0])
        self.assertEqual("", spark.sql("select trim_all('  ')").collect()[0][0])


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Dict, List, Sequence

import regex
from sqlfluff.core import Lexer, Linter, Parser, SQLBaseError
from sqlfluff.core.config import FluffConfig
from sqlfluff.core.parser import CodeSegment, RegexLexer

from easy_sql.sql_linter.rules import __all__
from easy_sql.sql_linter.sql_linter_reportor import LintReporter
from easy_sql.sql_processor.funcs import FuncRunner
from easy_sql.sql_processor.report import SqlProcessorReporter
from easy_sql.sql_processor.step import Step, StepFactory

if TYPE_CHECKING:
    from sqlfluff.core.parser.segments import BaseSegment


class SqlLinter:
    def __init__(self, sql: str, include_rules: List[str] = None, exclude_rules: List[str] = None):
        self.origin_sql = sql
        self.fixed_sql_list = []
        self.supported_backend = ["spark", "postgres", "clickhouse", "bigquery"]
        self.step_list = self._get_step_list()
        self.include_rules = include_rules
        self.exclude_rules = exclude_rules
        self.reporter = LintReporter()

    def _parse_backend(self, sql: str):
        sql_lines = sql.split("\n")
        backend = None
        for line in sql_lines:
            if re.match(r"^-- \s*backend:.*$", line):
                backend = line[line.index("backend:") + len("backend:") :].strip()
                break

        if backend is None:
            backend = "spark"
            self.reporter.report_warning("Backend cannot be found in sql, will use default backend spark")

        if backend not in self.supported_backend:
            raise Exception(
                f"Unsupported backend `${backend}`, all supported backends are: " + ",".join(self.supported_backend)
            )

        self.reporter.report_message(f"Use backend: {backend}")
        self.fixed_sql_list.append("-- backend: " + backend)
        return backend

    def _get_step_list(self):
        reporter = SqlProcessorReporter(report_task_id="sql_linter")
        func_runner = FuncRunner()
        step_factory = StepFactory(reporter, func_runner)
        step_list = step_factory.create_from_sql(self.origin_sql)
        return step_list

    def _get_dialect_from_backend(self, backend: str = None) -> str:
        backend = backend or self._parse_backend(self.origin_sql)
        if backend == "spark":
            return "sparksql"
        if backend == "bigquery":
            return "bigquery"
        if backend == "clickhouse":
            # TODO: so far do not have clickhouse in sql fluff
            return "ansi"
        if backend == "postgres":
            return "postgres"
        raise Exception("backend type so far is not supported for lint check")

    @staticmethod
    def _update_included_rule_for_config(config: Dict[str, Any], dialect: str = None, rules=None):
        if rules is None:
            rules = []
        if len(rules) > 0:
            config["core"]["rules"] = ",".join(rules)
        else:
            if dialect in ["bigquery"]:
                config["core"]["rules"] = "core," + dialect
            else:
                config["core"]["rules"] = "core"

    @staticmethod
    def _update_excluded_rule_for_config(config: Dict[str, Any], rules: List[str] = None):
        if rules is not None:
            config["core"]["exclude_rules"] = ",".join(rules)
        else:
            config["core"]["exclude_rules"] = None

    def _check_parsable(self, root_segment: BaseSegment) -> bool:
        segment_list = [root_segment]
        while len(segment_list) > 0:
            check_segment = segment_list[0]
            segment_list.remove(check_segment)
            if check_segment.is_type("unparsable"):
                self.reporter.report_warning("Unparsable segment found: " + check_segment.raw)
                return False
            elif check_segment.segments:
                segment_list = segment_list + list(check_segment.segments)
        return True

    def _check_lexable(self, tokens: Sequence[BaseSegment]) -> bool:
        for _, token in enumerate(tokens):
            if token.is_type("unlexable"):
                self.reporter.report_warning("Unlexable segment found: " + str(token.raw_segments))
                return False
        return True

    def _lint_step_sql(
        self, step: Step, lexer: Lexer, parser: Parser, linter: Linter, log_error: bool
    ) -> List[SQLBaseError]:
        self.fixed_sql_list.append(step.target_config.step_config_str)
        if step.select_sql:
            if step.is_template_statement():
                self.fixed_sql_list.append(step.select_sql + "\n")
                self.reporter.report_message("Skip template sql for this step.")
            else:
                sql = step.select_sql + "\n"
                tokens, _ = lexer.lex(sql)
                parsed = parser.parse(tokens)

                if self._check_lexable(tokens) and self._check_parsable(parsed):
                    result = linter.lint(parsed)
                    if log_error:
                        self.reporter.report_list_of_violations(result, step.target_config.line_no)
                    fixed_tree, violation = linter.fix(parsed)
                    self.fixed_sql_list.append(fixed_tree.raw)
                    return result
                else:
                    self.fixed_sql_list.append(step.select_sql)

    def _prepare_lexer(self, dialect: str):
        lexer = Lexer(dialect=dialect)

        # Add regex mather to lexer to enable easy sql
        easy_sql_function = RegexLexer("easy_sql_function", r"\${[^\s,]+\(.+\)}", CodeSegment)
        easy_sql_variable = RegexLexer("easy_sql_variable", r"\${[^\s,]+}", CodeSegment)
        easy_sql_template = RegexLexer("easy_sql_template", r"@{[^{]+}", CodeSegment)
        three_quote_string = RegexLexer("three_quote_string", r'""".*"""', CodeSegment)
        lexer.lexer_matchers.insert(0, easy_sql_variable)
        lexer.lexer_matchers.insert(0, easy_sql_function)
        lexer.lexer_matchers.insert(0, easy_sql_template)
        lexer.lexer_matchers.insert(0, three_quote_string)
        return lexer

    def _prepare_parser(self, dialect: str):
        parser = Parser(dialect=dialect)
        # Let parser recognize all function/variables/template into nakedIdentifier
        # which will be given grammar when given context
        identifier_segement = parser.config.get("dialect_obj")._library["NakedIdentifierSegment"]
        template = identifier_segement.template + r"|@{[^{]+}|[\$]{[\s\S]+}|\"[\s\S]+\""
        identifier_segement.template = template
        # For sqlfluff > 1.0, template is compiled and cached, so we need to update it as well.
        if identifier_segement._template is not None:
            identifier_segement._template = regex.compile(template, regex.IGNORECASE)
        return parser

    def _lint_for_easy_sql(self, backend: str, log_error: bool = True, config_path: str = None) -> List[SQLBaseError]:
        lint_result = []
        dialect = self._get_dialect_from_backend(backend)
        lexer = self._prepare_lexer(dialect)
        parser = self._prepare_parser(dialect)
        linter = self._prepare_linter(dialect, config_path)
        step_count = 0
        self.fixed_sql_list = self._parser_sql_header()
        for step_no, step in enumerate(self.step_list):
            step_count = step_count + 1
            if log_error:
                self.reporter.report_message(
                    "=== Check step {} at line {} ===".format(step_no + 1, step.target_config.line_no)
                )
            step_result = self._lint_step_sql(step, lexer, parser, linter, log_error)
            if step_result:
                lint_result = lint_result + step_result
        return lint_result

    def _lint_for_normal_sql(self, backend: str, log_error: bool = True, config_path: str = None) -> List[SQLBaseError]:
        self.fixed_sql_list = []
        dialect = self._get_dialect_from_backend(backend)
        lexer = Lexer(dialect=dialect)
        parser = Parser(dialect=dialect)
        linter = self._prepare_linter(dialect, config_path=config_path)
        tokens, _ = lexer.lex(self.origin_sql)
        parsed = parser.parse(tokens)
        lint_result = linter.lint(parsed)
        fixed_tree, violation = linter.fix(parsed)
        if log_error:
            self.reporter.report_list_of_violations(lint_result)
        self.fixed_sql_list.append(fixed_tree.raw)
        return lint_result

    def _parser_sql_header(self):
        line_no = self.step_list[0].target_config.line_no
        return self.origin_sql.split("\n")[: line_no - 1]

    def _prepare_linter(self, dialect: str, config_path: str = None):
        if config_path:
            config = FluffConfig.from_path(config_path)
        else:
            default_config_dict = FluffConfig(require_dialect=False)._configs
            default_config_dict["rules"]["L019"] = {"comma_style": "leading"}
            default_config_dict["rules"]["L014"] = {"extended_capitalisation_policy": "lower"}
            default_config_dict["core"]["dialect"] = dialect
            self._update_included_rule_for_config(default_config_dict, dialect=dialect, rules=self.include_rules)
            self._update_excluded_rule_for_config(default_config_dict, rules=self.exclude_rules)
            config = FluffConfig(configs=default_config_dict)
        linter = Linter(config=config, user_rules=__all__)
        return linter

    def lint(
        self, backend: str, log_error: bool = True, easysql: bool = True, config_path: str = None
    ) -> List[SQLBaseError]:
        if easysql:
            return self._lint_for_easy_sql(backend, log_error, config_path=config_path)
        else:
            return self._lint_for_normal_sql(backend, log_error, config_path=config_path)

    def fix(self, backend: str, log_linter_error: bool = False, easy_sql: bool = True, config_path: str = None):
        self.lint(backend, log_linter_error, easy_sql, config_path=config_path)
        return "\n".join(self.fixed_sql_list)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Dict, List, Tuple, Union

from easy_sql.sql_processor.backend.sql_dialect import SqlDialect

if TYPE_CHECKING:
    from sqlalchemy.engine import ResultProxy

    from ..base import Partition

__all__ = ["PgSqlDialect"]

from ...common import SqlProcessorAssertionError


class PostgrePartition:
    def __init__(self, name: str, value: Union[int, str]):
        self.name = name
        self.value = value
        self.value_next = value + "_" if isinstance(value, str) else value + 1
        self.value_expr = f"'{value}'" if isinstance(value, str) else str(value)
        self.value_next_expr = f"'{self.value_next}'" if isinstance(self.value_next, str) else str(self.value_next)
        self.table_suffix = str(value).lower().replace("-", "_")

    @property
    def field_name(self):
        return self.name

    def partition_table_name(self, table_name: str):
        return f"{table_name}__{self.table_suffix}"


class PgSqlDialect(SqlDialect):
    def create_db_sql(self, db: str) -> str:
        return f"create schema if not exists {db}"

    def use_db_sql(self, db: str) -> str:
        return f"set search_path='{db}'"

    def drop_db_sql(self, db: str):
        return f"drop schema if exists {db} cascade"

    def rename_table_sql(self, from_table: str, to_table: str) -> str:
        pure_to_table = to_table if "." not in to_table else to_table[to_table.index(".") + 1 :]
        return f"alter table {from_table} rename to {pure_to_table}"

    def rename_table_db_sql(self, table_name: str, schema: str):
        return f"alter table {table_name} set schema {schema}"

    def rename_view_sql(self, from_table: str, to_table: str) -> str:
        pure_to_table = to_table if "." not in to_table else to_table[to_table.index(".") + 1 :]
        return f"alter view {from_table} rename to {pure_to_table}"

    def create_view_sql(self, table_name: str, select_sql: str) -> str:
        return f"create view {table_name} as {select_sql}"

    def drop_view_sql(self, table_name: str) -> str:
        return f"drop view {table_name} cascade"

    def get_tables_sql(self, db) -> str:
        return (
            f"select tablename FROM pg_catalog.pg_tables where schemaname='{db}' union "
            f"select viewname FROM pg_catalog.pg_views where schemaname='{db}'"
        )

    def get_dbs_sql(self) -> str:
        return "select schemaname from pg_catalog.pg_tables union select schemaname FROM pg_catalog.pg_views"

    def create_table_sql(self, table_name: str, select_sql: str) -> str:
        return f"create table {table_name} as {select_sql}"

    def support_native_partition(self) -> bool:
        return True

    def support_move_individual_partition(self) -> bool:
        return True

    def delete_partition_sql(self, table_name, partitions: List[Partition]) -> str:
        if len(partitions) > 1:
            raise SqlProcessorAssertionError(
                f"Only support exactly one partition column, found: {[str(p) for p in partitions]}"
            )
        partition = PostgrePartition(partitions[0].field, partitions[0].value)
        return f"drop table if exists {partition.partition_table_name(table_name)}"

    def native_partitions_sql(self, table_name: str) -> Tuple[str, Callable[[ResultProxy], List[str]]]:
        if "." not in table_name:
            raise SqlProcessorAssertionError(
                "table name must be of format {DB}.{TABLE} when query native partitions, found: " + table_name
            )

        db, table = table_name[: table_name.index(".")], table_name[table_name.index(".") + 1 :]
        sql = f"""
        SELECT pg_catalog.pg_get_partkeydef(pt_table.oid) as partition_key
        FROM pg_class pt_table
            JOIN pg_namespace nmsp_pt_table   ON nmsp_pt_table.oid  = pt_table.relnamespace
        WHERE nmsp_pt_table.nspname='{db}' and pt_table.relname='{table}'
        """
        return sql, self.extract_partition_cols

    def extract_partition_cols(self, native_partitions_sql_result: ResultProxy) -> List[str]:
        partition_values = [v[0] for v in native_partitions_sql_result.fetchall()]
        if not partition_values:
            raise Exception("no partition values found, table may not exist")
        partition_value = partition_values[0]
        if partition_value is None:
            return []
        partition_value: str = partition_value
        if not partition_value.upper().startswith("RANGE (") or not partition_value.endswith(")"):
            raise Exception("unable to parse partition: " + partition_value)
        return partition_value[len("RANGE (") : -1].split(",")

    def create_table_with_partitions_sql(self, table_name: str, cols: List[Dict], partitions: List[Partition]):
        if len(partitions) > 1:
            raise SqlProcessorAssertionError(
                f"Only support exactly one partition column, found: {[str(p) for p in partitions]}"
            )
        cols_expr = ",\n".join([f"{col['name']} {col['type']}" for col in cols])
        partition_expr = f"partition by range({partitions[0].field})" if len(partitions) == 1 else ""
        return f"create table {table_name} (\n{cols_expr}\n) {partition_expr}"

    def create_partition_automatically(self):
        return False

    def create_temp_table_schema_like_target_schema(self):
        return False

    def create_partition_sql(
        self, target_table_name: str, partitions: List[Partition], if_not_exists: bool = False
    ) -> str:
        if len(partitions) > 1:
            raise SqlProcessorAssertionError(
                f"Only support exactly one partition column, found: {[str(p) for p in partitions]}"
            )
        partition = PostgrePartition(partitions[0].field, partitions[0].value)
        partition_table_name = partition.partition_table_name(target_table_name)
        if_not_exists = "if not exists" if if_not_exists else ""
        return (
            f"create table {if_not_exists} {partition_table_name} partition of {target_table_name} "
            f"for values from ({partition.value_expr}) to ({partition.value_next_expr})"
        )

    def create_partitions_with_data_sqls(
        self, source_table_name: str, target_table_name: str, col_names: List[str], partitions: List[List[Partition]]
    ):
        col_names_expr = ", ".join(col_names)
        if not partitions:
            return [
                f"insert into {target_table_name}({col_names_expr}) select {col_names_expr} from {source_table_name}"
            ]
        sqls = []
        for partition in partitions:
            if len(partition) > 1:
                raise SqlProcessorAssertionError(
                    f"Only support exactly one partition column, found: {[str(p) for p in partitions]}"
                )
            partition = PostgrePartition(partition[0].field, partition[0].value)
            partition_table_name = partition.partition_table_name(target_table_name)
            temp_table_name = f"{partition_table_name}__temp"
            sqls.append(f"drop table if exists {temp_table_name}")
            sqls.append(
                f"create table {temp_table_name} (like {target_table_name} including defaults including constraints)"
            )
            sqls.append(
                f"alter table {temp_table_name} add constraint p check ({partition.field_name} >="
                f" {partition.value_expr} and {partition.field_name} < {partition.value_next_expr})"
            )
            filter_expr = f"{partition.field_name} = {partition.value_expr}"
            sqls.append(
                f"insert into {temp_table_name}({col_names_expr}) select {col_names_expr} from"
                f" {source_table_name} where {filter_expr}"
            )
            sqls.append(f"drop table if exists {partition_table_name}")
            sqls.append(
                f'alter table {temp_table_name} rename to {partition_table_name[partition_table_name.index(".") + 1:]}'
            )
            sqls.append(
                f"alter table {target_table_name} attach partition {partition_table_name} "
                f"for values from ({partition.value_expr}) to ({partition.value_next_expr})"
            )
            sqls.append(f"alter table {partition_table_name} drop constraint p")
        return sqls

    def insert_data_sql(self, table_name: str, col_names_expr: str, select_sql: str, partitions: List[Partition]):
        return f"insert into {table_name}({col_names_expr}) {select_sql}"

    def move_data_sql(self, target_table: str, temp_table: str, partitions: List[Partition]) -> List[str]:
        sql = []
        for partition in partitions:
            pg_partition = PostgrePartition(partition.field, partition.value)
            temp_partition_table_name = pg_partition.partition_table_name(temp_table)
            target_partition_table_name = pg_partition.partition_table_name(target_table)
            sql.append(self.drop_table_sql(target_partition_table_name))
            sql.append(f"alter table {temp_table} detach partition {temp_partition_table_name}")
            sql.append((self.rename_table_sql(temp_partition_table_name, target_partition_table_name)))
            sql.append(
                f"alter table {target_table} attach partition {target_partition_table_name} "
                f"for values from ({pg_partition.value_expr}) to ({pg_partition.value_next_expr})"
            )
        return sql

    def drop_table_sql(self, table: str):
        return f"drop table if exists {table}"

    def support_static_partition(self):
        return True

    def create_table_like_sql(self, target_table_name: str, source_table_name: str, partitions: List[Partition]) -> str:
        if len(partitions) == 0:
            return f"create table {target_table_name} (like {source_table_name} including constraints)"
        elif len(partitions) > 1:
            raise SqlProcessorAssertionError(
                f"Only support exactly one partition column, found: {[str(p) for p in partitions]}"
            )
        else:
            partition_expr = f"partition by range({partitions[0].field})"
            return f"create table {target_table_name} (like {source_table_name} including constraints) {partition_expr}"



import unittest

from . import sql_test


class SqlTestTest(unittest.TestCase):
    def test_convert_test_for_spark(self):
        self.run_test("spark")

    def test_convert_test_for_postgres(self):
        self.run_test("postgres")

    def test_convert_test_for_clickhouse(self):
        self.run_test("clickhouse")

    def run_test(self, backend: str):
        sql_test._convert_json(f"test/sample_etl.{backend}.xlsx")
        sql_test._run_test(f"test/sample_etl.{backend}.xlsx", backend=backend)
        sql_test._run_test(f"test/sample_etl.{backend}.json", backend=backend)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Dict, List, Tuple

from easy_sql.sql_processor.backend.sql_dialect import SqlDialect, SqlExpr

if TYPE_CHECKING:
    from sqlalchemy.engine import ResultProxy
    from sqlalchemy.types import TypeEngine

    from ..base import Partition

__all__ = ["BqSqlDialect"]

from ...common import SqlProcessorAssertionError


class BqSqlDialect(SqlDialect):
    def __init__(self, db: str, sql_expr: SqlExpr):
        self.db = db
        super().__init__(sql_expr)

    def create_db_sql(self, db: str) -> str:
        return f"create schema if not exists {db}"

    # There's no such equivalent statement like 'use ${db}' in BigQuery.
    # Table must be qualified with a dataset when using BigQuery
    def use_db_sql(self, db: str) -> str:
        return "select 1"

    def drop_db_sql(self, db: str) -> str:
        return f"drop schema if exists {db} cascade"

    def rename_table_sql(self, from_table: str, to_table: str) -> str:
        from_full_table_name = from_table if self.contains_db(from_table) else f"{self.db}.{from_table}"
        to_pure_table_name = to_table.split(".")[1] if self.contains_db(to_table) else to_table
        return f"alter table if exists {from_full_table_name} rename to {to_pure_table_name}"

    # There's no statement that could change the dataset directly
    # Copy cannot be operated at view
    def rename_table_db_sql(self, table_name: str, schema: str):
        pure_table_name = table_name[table_name.index(".") + 1 :] if self.contains_db(table_name) else table_name
        from_table_name = table_name if self.contains_db(table_name) else f"{self.db}.{table_name}"
        return f"create table if not exists {schema}.{pure_table_name} copy {from_table_name}"

    def get_tables_sql(self, db) -> str:
        return f"select table_name from {db}.INFORMATION_SCHEMA.TABLES"

    def create_table_sql(self, table_name: str, select_sql: str) -> str:
        full_table_name = table_name if self.contains_db(table_name) else f"{self.db}.{table_name}"
        return f"create table if not exists {full_table_name} as {select_sql}"

    # Cannot rename view directly in BigQuery
    def rename_view_sql(self, from_table: str, to_table: str) -> str:
        from_table_name = from_table if self.contains_db(from_table) else f"{self.db}.{from_table}"
        to_table_name = to_table if self.contains_db(to_table) else f"{self.db}.{to_table}"
        return f"create view if not exists {to_table_name} as select * from {from_table_name}"

    def drop_view_sql(self, table: str) -> str:
        return f"drop view if exists {self.db}.{table}"

    def create_view_sql(self, table_name: str, select_sql: str) -> str:
        full_table_name = table_name if self.contains_db(table_name) else f"{self.db}.{table_name}"
        return f"create view if not exists {full_table_name} as {select_sql}"

    def support_native_partition(self) -> bool:
        return True

    def delete_partition_sql(self, table_name, partitions: List[Partition]) -> str:
        if not self.contains_db(table_name):
            raise SqlProcessorAssertionError("BigQuery table must be qualified with a dataset.")

        db, pure_table_name = tuple(table_name.split("."))
        if not partitions or len(partitions) == 0:
            raise Exception(
                f"cannot delete partition when partitions not specified: table_name={table_name},"
                f" partitions={partitions}"
            )
        if len(partitions) > 1:
            raise SqlProcessorAssertionError("BigQuery only supports single-column partitioning.")
        pt_expr = f"""{partitions[0].field} = '{partitions[0].value}'"""

        delete_pt_sql = f"delete {db}.{pure_table_name} where {pt_expr};"
        delete_pt_metadata = (
            f"delete {db}.__table_partitions__ where table_name = '{pure_table_name}' and partition_value ="
            f" '{partitions[0].value}';"
        )
        return self.transaction(f"{delete_pt_sql}\n{delete_pt_metadata}")

    def native_partitions_sql(self, table_name: str) -> Tuple[str, Callable[[ResultProxy], List[str]]]:
        db = table_name.split(".")[0] if self.contains_db(table_name) else {self.db}
        pure_table_name = table_name.split(".")[1] if self.contains_db(table_name) else {table_name}
        return (
            f"select ddl from {db}.INFORMATION_SCHEMA.TABLES where table_name = '{pure_table_name}'",
            self.extract_partition_cols,
        )

    def extract_partition_cols(self, native_partitions_sql_result):
        create_table_sql_lines = native_partitions_sql_result.fetchall()[0][0].split("\n")
        pt_cols = []
        for line in create_table_sql_lines:
            if line.startswith("PARTITION BY "):
                pt_cols = [line[len("PARTITION BY ") : -1].strip()]
                break
        return pt_cols

    def create_table_with_partitions_sql(
        self, table_name: str, cols: List[Dict[str, TypeEngine]], partitions: List[Partition]
    ):
        cols_expr = ",\n".join(
            f"{col['name']} {self.sql_expr.for_bigquery_type(col['name'], col['type'])}" for col in cols
        )
        if len(partitions) == 0:
            partition_expr = ""
        elif len(partitions) == 1:
            partition_expr = f"partition by {self.sql_expr.bigquery_partition_expr(partitions[0].field)}"
        else:
            raise SqlProcessorAssertionError("BigQuery only supports single-column partitioning.")
        table_name_with_db = table_name if self.contains_db(table_name) else f"{self.db}.{table_name}"
        return f"create table if not exists {table_name_with_db} (\n{cols_expr}\n)\n{partition_expr}\n"

    def create_partition_automatically(self):
        return True

    def insert_data_sql(self, table_name: str, col_names_expr: str, select_sql: str, partitions: List[Partition]):
        if not self.contains_db(table_name):
            raise SqlProcessorAssertionError("BigQuery table must be qualified with a dataset.")
        if any([pt.value is None for pt in partitions]):
            raise SqlProcessorAssertionError(
                f"cannot insert data when partition value is None, partitions: {partitions}, "
                "there maybe some bug, please check"
            )

        insert_date_sql = f"insert into {table_name}({col_names_expr}) {select_sql};"
        delete_pt_metadata_if_exist = self.delete_pt_metadata_sql(table_name, partitions)
        insert_pt_metadata = self.insert_pt_metadata_sql(table_name, partitions)
        return self.transaction(f"{insert_date_sql}\n{delete_pt_metadata_if_exist}\n{insert_pt_metadata}")

    def drop_table_sql(self, table: str):
        if not self.contains_db(table):
            raise SqlProcessorAssertionError("BigQuery table must be qualified with a dataset.")
        db, pure_table_name = tuple(table.split("."))
        drop_table_sql = f"drop table if exists {db}.{pure_table_name};"
        drop_pt_metadata = f"delete {db}.__table_partitions__ where table_name = '{pure_table_name}';"
        return f"{drop_table_sql}\n{drop_pt_metadata}"

    def support_static_partition(self):
        return False

    def create_pt_meta_table_sql(self, db: str) -> str:
        return f"""create table if not exists {db}.__table_partitions__(
        table_name string, partition_value string, last_modified_time timestamp)
        cluster by table_name;"""

    def insert_pt_metadata_sql(self, table_name: str, partitions: List[Partition]) -> str:
        if len(partitions) == 0:
            return ""
        elif len(partitions) > 1:
            raise SqlProcessorAssertionError("BigQuery only supports single-column partitioning.")
        else:
            if not self.contains_db(table_name):
                raise SqlProcessorAssertionError("BigQuery table must be qualified with a dataset.")
            db, pure_table_name = tuple(table_name.split("."))
            return (
                f"insert into {db}.__table_partitions__ values ('{pure_table_name}', '{partitions[0].value}',"
                " CURRENT_TIMESTAMP());"
            )

    def delete_pt_metadata_sql(self, table_name: str, partitions: List[Partition]) -> str:
        if len(partitions) == 0:
            return ""
        elif len(partitions) > 1:
            raise SqlProcessorAssertionError("BigQuery only supports single-column partitioning.")
        else:
            if not self.contains_db(table_name):
                raise SqlProcessorAssertionError("BigQuery table must be qualified with a dataset.")
            db, pure_table_name = tuple(table_name.split("."))
            return (
                f"delete {db}.__table_partitions__ where table_name = '{pure_table_name}' and partition_value ="
                f" '{partitions[0].value}';"
            )

    def create_table_like_sql(self, target_table_name: str, source_table_name: str, partitions: List[Partition]) -> str:
        return f"create table {target_table_name} like {source_table_name}"

    @staticmethod
    def transaction(statement: str) -> str:
        return f"BEGIN TRANSACTION;\n{statement}\nCOMMIT TRANSACTION;"

    @staticmethod
    def contains_db(table_name: str) -> bool:
        return "." in table_name
